from typing import Tuple
from typing import TypeVar, Sequence
from typing import Union

from lazy import lazy
from nonebot import on_regex, Bot
from nonebot.internal.adapter import Event
from nonebot.internal.matcher import Matcher
from nonebot.internal.params import Depends
from nonebot.typing import T_State

from nonebot_plugin_pixivbot.enums import RankingMode
from nonebot_plugin_pixivbot.protocol_dep.post_dest import PostDestination
from nonebot_plugin_pixivbot.utils.decode_integer import decode_integer
from nonebot_plugin_pixivbot.utils.errors import BadRequestError
from .base import CommonHandler
from ..entry_handler import post_destination
from ..pkg_context import context
from ..utils import get_common_query_rule

UID = TypeVar("UID")
GID = TypeVar("GID")


@context.root.register_eager_singleton()
class RankingHandler(CommonHandler):
    @classmethod
    def type(cls) -> str:
        return "ranking"

    mode_mapping = {RankingMode.day: "日", RankingMode.week: "周", RankingMode.month: "月",
                    RankingMode.day_male: "男性", RankingMode.day_female: "女性", RankingMode.week_original: "原创",
                    RankingMode.week_rookie: "新人", RankingMode.day_manga: "漫画"}

    mode_rev_mapping = {}
    for mode, text in mode_mapping.items():
        mode_rev_mapping[text] = mode

    def enabled(self) -> bool:
        return self.conf.pixiv_ranking_query_enabled

    @lazy
    def matcher(self):
        return on_regex(r"^看看(.*)?榜\s*(.*)?$", rule=get_common_query_rule(), priority=4, block=True)

    async def on_match(self, bot: Bot, event: Event, state: T_State, matcher: Matcher,
                       post_dest: PostDestination[UID, GID] = Depends(post_destination)):
        if "_matched_groups" in state:
            mode = state["_matched_groups"][0]
            num = state["_matched_groups"][1]
        else:
            mode = None
            num = None
        await self.handle(mode, num, post_dest=post_dest)

    def validate_range(self, range: Tuple[int, int] = None):
        if range:
            start, end = range
            if end - start + 1 > self.conf.pixiv_ranking_max_item_per_query:
                raise BadRequestError(
                    f"仅支持一次查询{self.conf.pixiv_ranking_max_item_per_query}张以下插画")
            elif start > end:
                raise BadRequestError("范围不合法")
            elif end > self.conf.pixiv_ranking_fetch_item:
                raise BadRequestError(
                    f'仅支持查询{self.conf.pixiv_ranking_fetch_item}名以内的插画')

    def parse_args(self, args: Sequence[str], post_dest: PostDestination[UID, GID]) -> dict:
        mode = args[0] if len(args) > 0 else None
        range = args[1] if len(args) > 1 else None

        if not mode:  # 判断是不是空字符串
            mode = None
        elif mode in self.mode_rev_mapping:
            mode = self.mode_rev_mapping[mode]
        elif isinstance(mode, str):
            try:
                mode = RankingMode[mode]
            except:
                raise BadRequestError(f"{mode}不是合法的榜单类型")

        if not range:
            range = None
        else:
            try:
                if "-" in range:
                    start, end = range.split("-")
                    range = int(start), int(end)
                else:
                    range = decode_integer(range)
                    range = range, range
            except ValueError:
                raise BadRequestError(f"{range}不是合法的范围")

        return {"mode": mode, "range": range}

    async def actual_handle(self, *, mode: Union[RankingMode, None] = None,
                            range: Union[Tuple[int, int], int, None] = None,
                            post_dest: PostDestination[UID, GID],
                            silently: bool = False):
        if mode is None:
            mode = self.conf.pixiv_ranking_default_mode

        if range is None:
            range = self.conf.pixiv_ranking_default_range

        if isinstance(range, int):
            range = range, range

        self.validate_range(range)
        illusts = await self.service.illust_ranking(mode, range)
        await self.post_illusts(illusts,
                                header=f"这是您点的{self.mode_mapping[mode]}榜",
                                number=range[0],
                                post_dest=post_dest)



from abc import ABC, abstractmethod
from functools import wraps
from typing import Callable, TypeVar

from nonebot_plugin_pixivbot import context
from nonebot_plugin_pixivbot.context import Inject
from nonebot_plugin_pixivbot.protocol_dep.post_dest import PostDestination
from nonebot_plugin_pixivbot.protocol_dep.postman import PostmanManager

UID = TypeVar("UID")
GID = TypeVar("GID")


@context.inject
class Interceptor(ABC):
    postman_manager = Inject(PostmanManager)

    async def post_plain_text(self, message: str,
                              post_dest: PostDestination):
        await self.postman_manager.send_plain_text(message, post_dest=post_dest)

    @abstractmethod
    async def intercept(self, wrapped_func: Callable, *args,
                        post_dest: PostDestination[UID, GID],
                        silently: bool,
                        **kwargs):
        raise NotImplementedError()

    def __call__(self, wrapped: Callable):
        @wraps(wrapped)
        async def wrapper(*args, post_dest: PostDestination[UID, GID],
                          silently: bool,
                          **kwargs):
            await self.intercept(wrapped, *args, post_dest=post_dest, silently=silently, **kwargs)

        return wrapper



from abc import ABC, abstractmethod
from typing import TypeVar, Union, Awaitable, Generic

from nonebot_plugin_pixivbot.global_context import context
from nonebot_plugin_pixivbot.protocol_dep.post_dest import PostDestination
from nonebot_plugin_pixivbot.protocol_dep.protocol_dep import ProtocolDep, ProtocolDepManager

UID = TypeVar("UID")
GID = TypeVar("GID")


class Authenticator(ProtocolDep, ABC, Generic[UID, GID]):
    @abstractmethod
    def group_admin(self, post_dest: PostDestination[UID, GID]) -> Union[bool, Awaitable[bool]]:
        raise NotImplementedError()

    @abstractmethod
    def available(self, post_dest: PostDestination[UID, GID]) -> Union[bool, Awaitable[bool]]:
        raise NotImplementedError()


@context.register_singleton()
class AuthenticatorManager(ProtocolDepManager[Authenticator]):
    def group_admin(self, post_dest: PostDestination[UID, GID]) -> Union[bool, Awaitable[bool]]:
        return self[post_dest.adapter].group_admin(post_dest)

    def available(self, post_dest: PostDestination[UID, GID]) -> Union[bool, Awaitable[bool]]:
        return self[post_dest.adapter].available(post_dest)


__all__ = ("Authenticator", "AuthenticatorManager")



from datetime import datetime, timezone
from enum import Enum
from typing import Any, TypeVar, Generic, Dict

from pydantic.generics import GenericModel

from nonebot_plugin_pixivbot.model import PostIdentifier

UID = TypeVar("UID")
GID = TypeVar("GID")


class WatchType(str, Enum):
    user_illusts = "user_illusts"
    following_illusts = "following_illusts"


class WatchTask(GenericModel, Generic[UID, GID]):
    code: int = 0
    type: WatchType
    kwargs: Dict[str, Any]
    subscriber: PostIdentifier[UID, GID]
    checkpoint: datetime = datetime.now(timezone.utc)

    class Config:
        orm_mode = True



from typing import TypeVar, Optional, Any

from beanie import Document
from beanie.odm.operators.update.general import Set
from pymongo import IndexModel

from nonebot_plugin_pixivbot.global_context import context
from nonebot_plugin_pixivbot.model import PixivBinding
from ..source import MongoDataSource

UID = TypeVar("UID")


class PixivBindingDocument(PixivBinding[Any], Document):
    class Settings:
        name = "pixiv_binding"
        indexes = [
            IndexModel([("adapter", 1), ("user_id", 1)], unique=True)
        ]


context.require(MongoDataSource).document_models.append(PixivBindingDocument)


@context.register_singleton()
class MongoPixivBindingRepo:
    async def get(self, adapter: str, user_id: UID) -> Optional[PixivBinding]:
        result = await PixivBindingDocument.find_one(
            PixivBindingDocument.adapter == adapter,
            PixivBindingDocument.user_id == user_id
        )
        return result

    async def update(self, binding: PixivBinding):
        await PixivBindingDocument.find_one(
            PixivBindingDocument.adapter == binding.adapter,
            PixivBindingDocument.user_id == binding.user_id
        ).upsert(
            Set({PixivBindingDocument.pixiv_user_id: binding.pixiv_user_id}),
            on_insert=PixivBindingDocument(**binding.dict())
        )

    async def remove(self, adapter: str, user_id: UID) -> bool:
        cnt = await PixivBindingDocument.find_one(
            PixivBindingDocument.adapter == adapter,
            PixivBindingDocument.user_id == user_id
        ).delete()
        return cnt == 1


__all__ = ("MongoPixivBindingRepo",)



import asyncio
import functools
import multiprocessing
from concurrent.futures.thread import ThreadPoolExecutor
from io import BytesIO

from PIL import Image, ImageFile

from nonebot_plugin_pixivbot import context
from nonebot_plugin_pixivbot.config import Config
from nonebot_plugin_pixivbot.context import Inject


@context.inject
@context.register_singleton()
class Compressor:
    _conf = Inject(Config)

    def __init__(self) -> None:
        self.enabled = self._conf.pixiv_compression_enabled
        self.max_size = self._conf.pixiv_compression_max_size
        self.quantity = self._conf.pixiv_compression_quantity

        if self.enabled:
            cpu_count = multiprocessing.cpu_count()
            self._executor = ThreadPoolExecutor(cpu_count, "compressor")
            # logger.info(f"A ThreadPool with {cpu_count} worker(s) was created for compression")

    async def compress(self, content: bytes) -> bytes:
        if self.enabled:
            loop = asyncio.get_running_loop()
            task = loop.run_in_executor(
                self._executor, functools.partial(self._compress, content))
            return await task
        else:
            return content

    def _compress(self, content: bytes) -> bytes:
        p = ImageFile.Parser()
        p.feed(content)
        img = p.close()

        w, h = img.size
        if w > self.max_size or h > self.max_size:
            ratio = min(self.max_size / w,
                        self.max_size / h)
            img_cp = img.resize(
                (int(ratio * w), int(ratio * h)), Image.ANTIALIAS)
        else:
            img_cp = img.copy()
        img_cp = img_cp.convert("RGB")

        with BytesIO() as bio:
            img_cp.save(bio, format="JPEG", optimize=True,
                        quantity=self.quantity)
            return bio.getvalue()



from typing import TypeVar, Optional, AsyncIterable, Collection

import tzlocal
from sqlalchemy import Column, Integer, Enum as SqlEnum, JSON, String, select, Index
from sqlalchemy.dialects.sqlite import insert

from nonebot_plugin_pixivbot import context
from nonebot_plugin_pixivbot.context import Inject
from nonebot_plugin_pixivbot.data.seq import SeqRepo
from nonebot_plugin_pixivbot.data.source.sql import SqlDataSource
from nonebot_plugin_pixivbot.data.utils.process_subscriber import process_subscriber
from nonebot_plugin_pixivbot.model import Subscription, PostIdentifier, ScheduleType

UID = TypeVar("UID")
GID = TypeVar("GID")

ID = PostIdentifier[UID, GID]


@context.require(SqlDataSource).registry.mapped
class SubscriptionOrm:
    __tablename__ = "subscription"

    id = Column(Integer, primary_key=True, autoincrement=True)
    subscriber = Column(JSON, nullable=False)
    code = Column(Integer, nullable=False)
    type = Column(SqlEnum(ScheduleType), nullable=False)
    kwargs = Column(JSON, nullable=False, default=dict)
    adapter = Column(String, nullable=False)
    schedule = Column(JSON, nullable=False)
    tz = Column(String, nullable=False, default=tzlocal.get_localzone_name)

    __table_args__ = (
        Index("adapter"),
        Index("subscriber", "code"),
    )


@context.inject
@context.register_singleton()
class SqlSubscriptionRepo:
    data_source: SqlDataSource = Inject(SqlDataSource)
    seq_repo: SeqRepo = Inject(SeqRepo)

    async def get_by_subscriber(self, subscriber: ID) -> AsyncIterable[Subscription]:
        subscriber = process_subscriber(subscriber)

        session = self.data_source.session()
        stmt = (
            select(SubscriptionOrm)
            .where(SubscriptionOrm.subscriber == subscriber.dict())
        )
        async for x in await session.stream_scalars(stmt):
            yield Subscription.from_orm(x)

    async def get_by_adapter(self, adapter: str) -> AsyncIterable[Subscription]:
        session = self.data_source.session()
        stmt = (
            select(SubscriptionOrm)
            .where(SubscriptionOrm.adapter == adapter)
        )
        async for x in await session.stream_scalars(stmt):
            yield Subscription.from_orm(x)

    async def insert(self, subscription: Subscription):
        subscription.subscriber = process_subscriber(subscription.subscriber)
        subscription.code = await self.seq_repo.inc_and_get(f'subscription {subscription.subscriber}')

        session = self.data_source.session()
        stmt = (insert(SubscriptionOrm)
                .values(subscriber=subscription.subscriber.dict(),
                        code=subscription.code,
                        type=subscription.type,
                        kwargs=subscription.kwargs,
                        adapter=subscription.subscriber.adapter,
                        schedule=subscription.schedule,
                        tz=subscription.tz))
        await session.execute(stmt)
        await session.commit()

    async def delete_one(self, subscriber: ID, code: int) -> Optional[Subscription]:
        subscriber = process_subscriber(subscriber)

        session = self.data_source.session()
        stmt = (select(SubscriptionOrm)
                .where(SubscriptionOrm.subscriber == subscriber.dict(),
                       SubscriptionOrm.code == code)
                .limit(1))
        sub = (await session.execute(stmt)).scalar_one_or_none()

        if sub is None:
            return sub

        await session.delete(sub)
        await session.commit()
        return Subscription.from_orm(sub)

    async def delete_many_by_subscriber(self, subscriber: ID) -> Collection[Subscription]:
        subscriber = process_subscriber(subscriber)

        session = self.data_source.session()
        stmt = (select(SubscriptionOrm)
                .where(SubscriptionOrm.subscriber == subscriber.dict()))
        subs = (await session.execute(stmt)).scalars().all()

        for t in subs:
            await session.delete(t)

        await session.commit()
        return subs


__all__ = ("SqlSubscriptionRepo",)



from datetime import datetime
from typing import Optional, List

from sqlalchemy import Column, JSON, Integer, BLOB, DateTime, ForeignKey, String, UniqueConstraint
from sqlalchemy.orm import relationship

from nonebot_plugin_pixivbot import context
from nonebot_plugin_pixivbot.data.source.sql import SqlDataSource


@context.require(SqlDataSource).registry.mapped
class DownloadCache:
    __tablename__ = "download_cache"

    illust_id: int = Column(Integer, primary_key=True, nullable=False,
                            sqlite_on_conflict_primary_key='REPLACE')
    content: bytes = Column(BLOB, nullable=False)

    update_time: datetime = Column(DateTime, nullable=False)


@context.require(SqlDataSource).registry.mapped
class IllustDetailCache:
    __tablename__ = "illust_detail_cache"

    illust_id: int = Column(Integer, primary_key=True, nullable=False,
                            sqlite_on_conflict_primary_key='REPLACE')
    illust: dict = Column(JSON, nullable=False)

    update_time: datetime = Column(DateTime, nullable=False)


@context.require(SqlDataSource).registry.mapped
class UserDetailCache:
    __tablename__ = "user_detail_cache"

    user_id: int = Column(Integer, primary_key=True, nullable=False,
                          sqlite_on_conflict_primary_key='REPLACE')
    user: dict = Column(JSON, nullable=False)

    update_time: datetime = Column(DateTime, nullable=False)


@context.require(SqlDataSource).registry.mapped
class IllustSetCache:
    __tablename__ = "illust_set_cache"

    id: int = Column(Integer, primary_key=True, autoincrement=True, nullable=False)
    cache_type: str = Column(String, nullable=False)
    key: dict = Column(JSON, nullable=False)

    update_time: datetime = Column(DateTime, nullable=False)
    pages: Optional[int] = Column(Integer)
    next_qs: Optional[dict] = Column(JSON)

    illust_id: List["IllustSetCacheIllust"] = relationship("IllustSetCacheIllust",
                                                           foreign_keys="IllustSetCacheIllust.cache_id",
                                                           cascade="save-update, delete",
                                                           passive_deletes=True)

    size: int = Column(Integer, nullable=False, default=0)

    __table_args__ = (
        UniqueConstraint('cache_type', 'key'),
    )


@context.require(SqlDataSource).registry.mapped
class IllustSetCacheIllust:
    __tablename__ = "illust_set_cache_illust"

    cache_id: int = Column(Integer, ForeignKey("illust_set_cache.id"), primary_key=True, nullable=False)
    illust_id: int = Column(Integer, primary_key=True, nullable=False)
    rank: int = Column(Integer, nullable=False, default=0)


@context.require(SqlDataSource).registry.mapped
class UserSetCache:
    __tablename__ = "user_set_cache"

    id: int = Column(Integer, primary_key=True, autoincrement=True, nullable=False)
    cache_type: str = Column(String, nullable=False)
    key: dict = Column(JSON, nullable=False)

    update_time: datetime = Column(DateTime, nullable=False)
    pages: Optional[int] = Column(Integer)
    next_qs: Optional[dict] = Column(JSON)

    illust_id: List["UserSetCacheUser"] = relationship("UserSetCacheUser",
                                                       foreign_keys="UserSetCacheUser.cache_id",
                                                       cascade="save-update, delete",
                                                       passive_deletes=True)

    __table_args__ = (
        UniqueConstraint('cache_type', 'key'),
    )


@context.require(SqlDataSource).registry.mapped
class UserSetCacheUser:
    __tablename__ = "user_set_cache_user"

    cache_id: int = Column(Integer, ForeignKey("user_set_cache.id"), primary_key=True, nullable=False)
    user_id: int = Column(Integer, primary_key=True, nullable=False)



from nonebot_plugin_pixivbot.context import Context


def asyncio_scheduler_provider(context: Context):
    # 改成register_lazy以后Scheduler不工作，不懂为啥
    from nonebot import require

    require("nonebot_plugin_apscheduler")

    from nonebot_plugin_apscheduler import scheduler
    from apscheduler.schedulers.asyncio import AsyncIOScheduler

    context.register(AsyncIOScheduler, scheduler)


def base_scheduler_provider(context: Context):
    from apscheduler.schedulers.base import BaseScheduler
    from apscheduler.schedulers.asyncio import AsyncIOScheduler
    context.bind(BaseScheduler, AsyncIOScheduler)


providers = [asyncio_scheduler_provider, base_scheduler_provider]


def provide(context: Context):
    for p in providers:
        p(context)


__all__ = ("provide",)



from typing import Protocol, List, Union

from nonebot_plugin_pixivbot import context
from nonebot_plugin_pixivbot.config import Config
from nonebot_plugin_pixivbot.data.pixiv_repo import LazyIllust
from nonebot_plugin_pixivbot.data.pixiv_repo.models import PixivRepoMetadata
from nonebot_plugin_pixivbot.enums import RankingMode, DataSourceType
from nonebot_plugin_pixivbot.model import Illust, User


class LocalPixivRepo(Protocol):
    async def update_illust_detail(self, illust: Illust, metadata: PixivRepoMetadata):
        ...

    async def update_user_detail(self, user: User, metadata: PixivRepoMetadata):
        ...

    async def invalidate_search_illust(self, word: str):
        ...

    async def append_search_illust(self, word: str,
                                   content: List[Union[Illust, LazyIllust]],
                                   metadata: PixivRepoMetadata) -> bool:
        ...

    async def invalidate_search_user(self, word: str):
        ...

    async def append_search_user(self, word: str, content: List[User],
                                 metadata: PixivRepoMetadata) -> bool:
        ...

    async def invalidate_user_illusts(self, user_id: int):
        ...

    async def append_user_illusts(self, user_id: int,
                                  content: List[Union[Illust, LazyIllust]],
                                  metadata: PixivRepoMetadata) -> bool:
        ...

    async def invalidate_user_bookmarks(self, user_id: int):
        ...

    async def append_user_bookmarks(self, user_id: int,
                                    content: List[Union[Illust, LazyIllust]],
                                    metadata: PixivRepoMetadata) -> bool:
        ...

    async def invalidate_recommended_illusts(self):
        ...

    async def append_recommended_illusts(self, content: List[Union[Illust, LazyIllust]],
                                         metadata: PixivRepoMetadata) -> bool:
        ...

    async def invalidate_related_illusts(self, illust_id: int):
        ...

    async def append_related_illusts(self, illust_id: int,
                                     content: List[Union[Illust, LazyIllust]],
                                     metadata: PixivRepoMetadata) -> bool:
        ...

    async def invalidate_illust_ranking(self, mode: RankingMode):
        ...

    async def append_illust_ranking(self, mode: RankingMode,
                                    content: List[Union[Illust, LazyIllust]],
                                    metadata: PixivRepoMetadata) -> bool:
        ...

    async def update_image(self, illust_id: int, content: bytes,
                           metadata: PixivRepoMetadata):
        ...

    async def invalidate_all(self):
        ...


conf = context.require(Config)
if conf.pixiv_data_source == DataSourceType.mongo:
    from .mongo import MongoPixivRepo

    context.bind(LocalPixivRepo, MongoPixivRepo)
else:
    from .sql import SqlPixivRepo

    context.bind(LocalPixivRepo, SqlPixivRepo)

__all__ = ("LocalPixivRepo",)



from abc import ABC, abstractmethod
from asyncio import Lock, create_task
from contextlib import AbstractContextManager
from datetime import datetime, timezone
from inspect import isawaitable
from types import TracebackType
from typing import Any, Awaitable, Generic, TypeVar, Callable, Union, AsyncGenerator, List, Type, Optional, Dict

from nonebot import logger

from nonebot_plugin_pixivbot.data.pixiv_repo.enums import CacheStrategy
from nonebot_plugin_pixivbot.utils.expires_lru_dict import ExpiresLruDict

T_ID = TypeVar("T_ID")
T_ITEM = TypeVar("T_ITEM")


class SharedAsyncGeneratorContextManager(AbstractContextManager, Generic[T_ITEM]):
    def __init__(self, origin: AsyncGenerator[T_ITEM, None],
                 *, on_each: Optional[Callable[[T_ITEM], Union[None, Awaitable[None]]]] = None,
                 on_error: Optional[Callable[[Exception], Union[None, Awaitable[None]]]] = None,
                 on_stop: Optional[Callable[[List[T_ITEM]], Union[None, Awaitable[None]]]] = None,
                 on_consumers_changed: Optional[Callable[["SharedAsyncGeneratorContextManager", int], None]] = None):
        super().__init__()
        self._origin = origin
        self._stopped = False  # whether origin has raised a StopIteration
        self._got_items = []  # items got from origin, used to replay
        self._got = 0  # count of items got from origin
        self._mutex = Lock()  # to solve race
        self._consumers = 0  # count of consumer
        self._on_each = on_each  # callback on origin yields
        self._on_error = on_error  # callback on origin throws
        self._on_stop = on_stop  # callback on origin stops
        self._on_consumers_changed = on_consumers_changed  # callback on a consumer enters or exits

    @property
    def consumers(self) -> int:
        return self._consumers

    async def _generator(self) -> AsyncGenerator[T_ITEM, None]:
        cur = 0
        while True:
            if cur < self._got:
                yield self._got_items[cur]
                cur += 1
            else:
                if self._stopped:
                    break

                async with self._mutex:
                    if self._stopped:
                        break

                    if cur == self._got:
                        try:
                            new_data = await self._origin.__anext__()
                        except StopAsyncIteration:
                            self._stopped = True

                            if self._on_stop:
                                ret = self._on_stop(self._got_items)
                                if isawaitable(ret):
                                    await ret

                            break
                        except Exception as e:
                            if self._on_error:
                                ret = self._on_error(e)
                                if isawaitable(ret):
                                    await ret
                            raise e

                        if self._on_each:
                            ret = self._on_each(new_data)
                            if isawaitable(ret):
                                await ret

                        self._got_items.append(new_data)
                        self._got += 1

                    yield self._got_items[cur]
                    cur += 1

    def __enter__(self) -> AsyncGenerator[T_ITEM, None]:
        self._consumers += 1

        if self._on_consumers_changed:
            self._on_consumers_changed(self, self._consumers)

        return self._generator()

    def __exit__(self, exc_type: Optional[Type[BaseException]],
                 exc_value: Optional[BaseException],
                 traceback: Optional[TracebackType]) -> None:
        self._consumers -= 1

        if self._on_consumers_changed:
            self._on_consumers_changed(self, self._consumers)

    async def aclose(self):
        return await self._origin.aclose()

    def close(self):
        create_task(self.aclose())


class SharedAsyncGeneratorManager(ABC, Generic[T_ID, T_ITEM]):
    log_tag = "shared_agen"

    def __init__(self):
        self._ctx_mgr = {}
        self._paused_ctx_mgr = ExpiresLruDict(1024, self._cleanup)
        self._expires_time: Dict[T_ID, datetime] = {}

    @abstractmethod
    def agen(self, identifier: T_ID,
             cache_strategy: CacheStrategy,
             **kwargs) -> AsyncGenerator[T_ITEM, None]:
        raise NotImplementedError()

    async def on_agen_next(self, identifier: T_ID, item: T_ITEM):
        pass

    async def on_agen_stop(self, identifier: T_ID, items: List[T_ITEM]):
        pass

    async def on_agen_error(self, identifier: T_ID, e: Exception):
        self.invalidate(identifier)

    def _cleanup(self, identifier: T_ID, ctx_mgr: SharedAsyncGeneratorContextManager[T_ITEM]):
        logger.debug(f"[{self.log_tag}] {identifier} was popped")
        ctx_mgr.close()
        if identifier in self._expires_time:
            del self._expires_time[identifier]

    def on_consumers_changed(self, identifier: T_ID,
                             ctx_mgr: SharedAsyncGeneratorContextManager[T_ITEM],
                             consumers: int):
        if identifier in self._paused_ctx_mgr and consumers > 0:
            logger.debug(f"[{self.log_tag}] {identifier} re-started")
            del self._paused_ctx_mgr[identifier]
            self._ctx_mgr[identifier] = ctx_mgr
        elif identifier in self._ctx_mgr and consumers == 0:
            del self._ctx_mgr[identifier]
            if identifier in self._expires_time:
                logger.debug(f"[{self.log_tag}] {identifier} paused")
                self._paused_ctx_mgr.add(identifier, ctx_mgr, self._expires_time[identifier])
            else:
                self._cleanup(identifier, ctx_mgr)
                logger.debug(f"[{self.log_tag}] {identifier} stopped")

    def invalidate(self, identifier: T_ID):
        if identifier in self._paused_ctx_mgr:
            logger.debug(f"[{self.log_tag}] {identifier} was invalidated from paused state")
            self._paused_ctx_mgr.pop(identifier).close()
        elif identifier in self._ctx_mgr:
            logger.debug(f"[{self.log_tag}] {identifier} was invalidated from running state")
            self._ctx_mgr.pop(identifier)

        if identifier in self._expires_time:
            del self._expires_time[identifier]

    def invalidate_all(self):
        keys = list(self._paused_ctx_mgr.keys())
        for identifier in keys:
            logger.debug(f"[{self.log_tag}] {identifier} was invalidated from paused state")
            self._paused_ctx_mgr.pop(identifier).close()

        for identifier in self._ctx_mgr:
            logger.debug(f"[{self.log_tag}] {identifier} was invalidated from running state")
            self._ctx_mgr.pop(identifier)

        self._expires_time.clear()

    def get_expires_time(self, identifier: T_ID) -> Optional[datetime]:
        return self._expires_time.get(identifier, None)

    def set_expires_time(self, identifier: T_ID, expires_time: datetime):
        # 即使当前没有该ctx_mgr也可以设置
        # 在调用invalidate或者被_paused_ctx_mgr弹出后从_expires_time中删除
        if expires_time <= datetime.now(timezone.utc):
            self.invalidate(identifier)
        elif identifier not in self._expires_time:
            self._expires_time[identifier] = expires_time
            logger.debug(f"[{self.log_tag}] {identifier} will expire at {expires_time}")
        elif self._expires_time[identifier] != expires_time:
            logger.warning(f"[{self.log_tag}] {identifier}'s expires time was already set")

    def get(self, identifier: Any,
            cache_strategy: CacheStrategy = CacheStrategy.NORMAL,
            **kwargs) \
            -> SharedAsyncGeneratorContextManager[T_ITEM]:
        if cache_strategy == CacheStrategy.FORCE_EXPIRATION:
            self.invalidate(identifier)

        if identifier in self._paused_ctx_mgr:
            return self._paused_ctx_mgr[identifier]
        elif identifier in self._ctx_mgr:
            return self._ctx_mgr[identifier]
        else:
            origin = self.agen(identifier, cache_strategy, **kwargs)
            self._ctx_mgr[identifier] = SharedAsyncGeneratorContextManager(
                origin=origin,
                on_each=lambda item: self.on_agen_next(identifier, item),
                on_stop=lambda items: self.on_agen_stop(identifier, items),
                on_error=lambda e: self.on_agen_error(identifier, e),
                on_consumers_changed=lambda ctx_mgr, consumers: self.on_consumers_changed(
                    identifier, ctx_mgr, consumers)
            )
            return self._ctx_mgr[identifier]



from nonebot_plugin_pixivbot.data.pixiv_repo.models import PixivRepoMetadata


class LocalPixivRepoError(RuntimeError):
    pass


class NoSuchItemError(LocalPixivRepoError):
    pass


class CacheExpiredError(LocalPixivRepoError):
    def __init__(self, metadata: PixivRepoMetadata):
        self.metadata = metadata




from beanie import Document
from pymongo import IndexModel, ReturnDocument

from nonebot_plugin_pixivbot import context
from nonebot_plugin_pixivbot.context import Inject
from nonebot_plugin_pixivbot.data.source import MongoDataSource


class Seq(Document):
    key: str
    value: int = 0

    class Settings:
        name = "seq"
        indexes = [
            IndexModel([("key", 1)], unique=True),
        ]


context.require(MongoDataSource).document_models.append(Seq)


@context.inject
@context.register_singleton()
class MongoSeqRepo:
    mongo = Inject(MongoDataSource)

    async def inc_and_get(self, key: str) -> int:
        seq = await self.mongo.db.seq.find_one_and_update({'key': key},
                                                          {'$inc': {'value': 1}},
                                                          upsert=True,
                                                          return_document=ReturnDocument.AFTER)
        return seq["value"]

    async def get_and_inc(self, key: str) -> int:
        seq = await self.mongo.db.seq.find_one_and_update({'key': key},
                                                          {'$inc': {'value': 1}},
                                                          upsert=True,
                                                          return_document=ReturnDocument.BEFORE)
        return seq["value"]

    async def get(self, key: str) -> int:
        seq = await self.mongo.db.seq.find_one({'key': key})
        value = None
        if seq is not None:
            value = seq.get("value")
        return value if value is not None else 0



from . import mongo_v1_to_v2
from . import mongo_v2_to_v3
from . import mongo_v3_to_v4
from . import mongo_v4_to_v5
from .mongo_migration_manager import MongoMigrationManager

__all__ = ("MongoMigrationManager",)



from time import time

import numpy as np

from nonebot_plugin_pixivbot.data.pixiv_repo import LazyIllust
from nonebot_plugin_pixivbot.enums import RandomIllustMethod


def uniform(illusts: list[LazyIllust]) -> np.ndarray:
    # 概率相等
    n = len(illusts)
    return np.ones(n) / n


def bookmark_proportion(illusts: list[LazyIllust]) -> np.ndarray:
    # 概率正比于书签数
    n = len(illusts)
    p = [0] * n

    for i, x in enumerate(illusts):
        if x.loaded:
            p[i] = x.content.total_bookmarks + 10  # 加10平滑
        else:
            p[i] = 0

    p = np.array(p)
    return p / np.sum(p)


def view_proportion(illusts: list[LazyIllust]) -> np.ndarray:
    # 概率正比于查看人数
    n = len(illusts)
    p = [0] * n

    for i, x in enumerate(illusts):
        if x.loaded:
            p[i] = x.content.total_view + 10  # 加10平滑
        else:
            p[i] = 0

    p = np.array(p)
    return p / np.sum(p)


def timedelta_proportion(illusts: list[LazyIllust]) -> np.ndarray:
    # 概率正比于 exp(归一化后的画像发布时间差)
    n = len(illusts)
    p = [0] * n

    now = time()
    min_p = 0
    for i, x in enumerate(illusts):
        if x.loaded:
            p[i] = x.create_date.timestamp() - now
            if p[i] < min_p:
                min_p = p[i]
        else:
            p[i] = float("-inf")

    p = np.array(p)
    p = p / min_p  # 归一化
    p = np.exp(p)
    return p / np.sum(p)


p_gen = {
    RandomIllustMethod.uniform: uniform,
    RandomIllustMethod.bookmark_proportion: bookmark_proportion,
    RandomIllustMethod.view_proportion: view_proportion,
    RandomIllustMethod.timedelta_proportion: timedelta_proportion,
}


def roulette(illusts: list[LazyIllust], random_method: RandomIllustMethod, k: int) -> list[LazyIllust]:
    if random_method not in p_gen:
        raise ValueError(f"illegal random_method: {random_method}")

    n = len(illusts)
    p = p_gen[random_method](illusts)
    # print([(illusts[i].content.create_date, p[i]) for i in range(n)])

    rng = np.random.default_rng()
    winners = rng.choice(n, k, False, p)
    return [illusts[c] for c in winners]



from abc import ABC

from nonebot_plugin_pixivbot.context import Inject
from nonebot_plugin_pixivbot.service.pixiv_service import PixivService
from ..entry_handler import EntryHandler
from ..interceptor.cooldown_interceptor import CooldownInterceptor
from ..interceptor.loading_prompt_interceptor import LoadingPromptInterceptor
from ..interceptor.record_req_interceptor import RecordReqInterceptor
from ..interceptor.timeout_interceptor import TimeoutInterceptor
from ..pkg_context import context


@context.inject
class CommonHandler(EntryHandler, ABC):
    service = Inject(PixivService)

    def __init__(self):
        super().__init__()
        self.add_interceptor(context.require(CooldownInterceptor))
        self.add_interceptor(context.require(TimeoutInterceptor))
        self.add_interceptor(context.require(LoadingPromptInterceptor))


class RecordCommonHandler(CommonHandler, ABC):
    def __init__(self):
        super().__init__()
        self.add_interceptor(context.require(RecordReqInterceptor))



from .bind import BindHandler, UnbindHandler
from .command import CommandHandler
from .help import HelpHandler
from .invalidate_cache import InvalidateCacheHandler
from .schedule import ScheduleHandler, UnscheduleHandler
from .watch import WatchHandler, UnwatchHandler

__all__ = ("CommandHandler",)



from typing import TypeVar, Sequence

from lazy import lazy
from nonebot import on_regex, Bot
from nonebot.internal.adapter import Event
from nonebot.internal.matcher import Matcher
from nonebot.internal.params import Depends
from nonebot.typing import T_State

from nonebot_plugin_pixivbot.context import Inject
from nonebot_plugin_pixivbot.protocol_dep.post_dest import PostDestination
from nonebot_plugin_pixivbot.utils.errors import BadRequestError
from .base import RecordCommonHandler
from ..entry_handler import post_destination
from ..pkg_context import context
from ..recorder import Recorder
from ..utils import get_common_query_rule

UID = TypeVar("UID")
GID = TypeVar("GID")


@context.inject
@context.root.register_eager_singleton()
class RandomRelatedIllustHandler(RecordCommonHandler):
    recorder = Inject(Recorder)

    @classmethod
    def type(cls) -> str:
        return "random_related_illust"

    def enabled(self) -> bool:
        return self.conf.pixiv_random_related_illust_query_enabled

    @lazy
    def matcher(self):
        return on_regex("^不够色$", rule=get_common_query_rule(), priority=1, block=True)

    async def on_match(self, bot: Bot, event: Event, state: T_State, matcher: Matcher,
                       post_dest: PostDestination[UID, GID] = Depends(post_destination)):
        await self.handle(post_dest=post_dest)

    def parse_args(self, args: Sequence[str], post_dest: PostDestination[UID, GID]) -> dict:
        illust_id = self.recorder.get_resp(post_dest.identifier)
        if not illust_id:
            raise BadRequestError("你还没有发送过请求")
        return {"illust_id": illust_id}

    async def actual_handle(self, *, illust_id: int,
                            count: int = 1,
                            post_dest: PostDestination[UID, GID],
                            silently: bool = False):
        illusts = await self.service.random_related_illust(illust_id, count=count)

        await self.post_illusts(illusts,
                                header=f"这是您点的[{illust_id}]的相关图片",
                                post_dest=post_dest)



from . import command
from . import common
from . import recorder
from . import sniffer
from .entry_handler import EntryHandler, DelegationEntryHandler
from .handler import Handler

__all__ = ("Handler", "EntryHandler", "DelegationEntryHandler")



from typing import TypeVar, Callable

from nonebot_plugin_pixivbot.context import Inject
from nonebot_plugin_pixivbot.handler.interceptor.base import Interceptor
from nonebot_plugin_pixivbot.handler.recorder import Recorder, Req
from nonebot_plugin_pixivbot.protocol_dep.post_dest import PostDestination
from ..pkg_context import context

UID = TypeVar("UID")
GID = TypeVar("GID")


@context.inject
@context.register_singleton()
class RecordReqInterceptor(Interceptor):
    recorder = Inject(Recorder)

    async def intercept(self, wrapped_func: Callable, *args,
                        post_dest: PostDestination[UID, GID],
                        silently: bool,
                        **kwargs):
        await wrapped_func(*args, post_dest=post_dest, silently=silently, **kwargs)
        self.recorder.record_req(Req(wrapped_func, *args, **kwargs), post_dest.identifier)



from abc import ABC, abstractmethod
from inspect import isawaitable
from typing import TypeVar, Callable, Union, Awaitable, Optional

from nonebot import get_driver, logger

from nonebot_plugin_pixivbot.config import Config
from nonebot_plugin_pixivbot.context import Inject
from nonebot_plugin_pixivbot.protocol_dep.authenticator import AuthenticatorManager
from nonebot_plugin_pixivbot.protocol_dep.post_dest import PostDestination
from .base import Interceptor
from ..pkg_context import context

UID = TypeVar("UID")
GID = TypeVar("GID")


class PermissionInterceptor(Interceptor, ABC):
    @abstractmethod
    def has_permission(self, post_dest: PostDestination[UID, GID]) -> Union[bool, Awaitable[bool]]:
        raise NotImplementedError()

    def get_permission_denied_msg(self, post_dest: PostDestination[UID, GID]) \
            -> Union[Optional[str], Awaitable[Optional[str]]]:
        return None

    async def intercept(self, wrapped_func: Callable, *args,
                        post_dest: PostDestination[UID, GID],
                        silently: bool,
                        **kwargs):
        p = self.has_permission(post_dest)
        if isawaitable(p):
            p = await p

        if p:
            await wrapped_func(*args, post_dest=post_dest, silently=silently, **kwargs)
        else:
            logger.debug(f"permission denied {post_dest}")
            if not silently:
                msg = self.get_permission_denied_msg(post_dest)
                if isawaitable(msg):
                    await msg
                if msg:
                    await self.post_plain_text(msg, post_dest=post_dest)


class AnyPermissionInterceptor(PermissionInterceptor):
    def __init__(self, *interceptors: PermissionInterceptor):
        super().__init__()
        self.interceptors = list(interceptors)

    def append(self, interceptor: PermissionInterceptor):
        self.interceptors.append(interceptor)

    async def has_permission(self, post_dest: PostDestination[UID, GID]) -> bool:
        for inter in self.interceptors:
            p = inter.has_permission(post_dest)
            if isawaitable(p):
                p = await p

            if p:
                return True

        return False


@context.register_singleton()
class SuperuserInterceptor(PermissionInterceptor):
    def __init__(self):
        super().__init__()
        self.superusers = get_driver().config.superusers

    def has_permission(self, post_dest: PostDestination[UID, GID]) -> bool:
        return str(post_dest.user_id) in self.superusers \
               or f"{post_dest.adapter}:{post_dest.user_id}" in self.superusers


@context.inject
@context.register_singleton()
class GroupAdminInterceptor(PermissionInterceptor):
    auth = Inject(AuthenticatorManager)

    def has_permission(self, post_dest: PostDestination[UID, GID]) -> Union[bool, Awaitable[bool]]:
        if not post_dest.group_id:
            return True
        return self.auth.group_admin(post_dest)


@context.inject
@context.register_singleton()
class BlacklistInterceptor(PermissionInterceptor):
    conf = Inject(Config)

    def __init__(self):
        super().__init__()
        self.blacklist = self.conf.blacklist

    def has_permission(self, post_dest: PostDestination[UID, GID]) -> bool:
        return str(post_dest.user_id) not in self.blacklist \
               and f"{post_dest.adapter}:{post_dest.user_id}" not in self.blacklist



from abc import ABC, abstractmethod
from typing import TypeVar, Type

from nonebot.adapters import Bot, Event
from nonebot.internal.params import Depends
from nonebot.matcher import Matcher
from nonebot.typing import T_State

from nonebot_plugin_pixivbot.config import Config
from nonebot_plugin_pixivbot.global_context import context
from nonebot_plugin_pixivbot.handler.handler import Handler
from .interceptor.default_error_interceptor import DefaultErrorInterceptor
from .interceptor.permission_interceptor import BlacklistInterceptor
from .interceptor.sql_remove_session_interceptor import SqlRemoveSessionInterceptor
from .pkg_context import context
from ..context import Inject
from ..enums import DataSourceType
from ..protocol_dep.post_dest import PostDestinationFactoryManager, PostDestination

UID = TypeVar("UID")
GID = TypeVar("GID")


def post_destination(bot: Bot, event: Event):
    return context.require(PostDestinationFactoryManager).from_event(bot, event)


@context.inject
class EntryHandler(Handler, ABC):
    conf: Config = Inject(Config)

    def __init__(self):
        super().__init__()
        self.add_interceptor(context.require(DefaultErrorInterceptor))
        self.add_interceptor(context.require(BlacklistInterceptor))

        if self.conf.pixiv_data_source == DataSourceType.sqlite:
            self.add_interceptor(context.require(SqlRemoveSessionInterceptor))

        self.matcher.append_handler(self.on_match)

    @property
    @abstractmethod
    def matcher(self) -> Type[Matcher]:
        raise NotImplementedError()

    @abstractmethod
    async def on_match(self, bot: Bot, event: Event, state: T_State, matcher: Matcher,
                       post_dest: PostDestination[UID, GID] = Depends(post_destination)):
        raise NotImplementedError()


class DelegationEntryHandler(EntryHandler, ABC):
    @property
    @abstractmethod
    def delegation(self) -> EntryHandler:
        raise NotImplementedError()

    async def handle(self, *args,
                     post_dest: PostDestination[UID, GID],
                     silently: bool = False,
                     disabled_interceptors: bool = False,
                     **kwargs):
        if not self.enabled():
            return

        await self.delegation.handle(*args, post_dest=post_dest, silently=silently,
                                     disabled_interceptors=disabled_interceptors, **kwargs)

    async def actual_handle(self, *args, **kwargs):
        raise RuntimeError("unexpected call here")



from datetime import datetime, timedelta, timezone
from inspect import isawaitable
from typing import Dict, Any, TypeVar, AsyncIterable

from apscheduler.schedulers.asyncio import AsyncIOScheduler
from apscheduler.triggers.interval import IntervalTrigger
from nonebot import logger, Bot
from nonebot.exception import ActionFailed

from nonebot_plugin_pixivbot import context
from nonebot_plugin_pixivbot.config import Config
from nonebot_plugin_pixivbot.context import Inject
from nonebot_plugin_pixivbot.data.watch_task import WatchTaskRepo
from nonebot_plugin_pixivbot.handler.watch.following_illusts import WatchFollowingIllustsHandler
from nonebot_plugin_pixivbot.handler.watch.user_illusts import WatchUserIllustsHandler
from nonebot_plugin_pixivbot.model import PostIdentifier, WatchTask, WatchType
from nonebot_plugin_pixivbot.protocol_dep.authenticator import AuthenticatorManager
from nonebot_plugin_pixivbot.protocol_dep.post_dest import PostDestination, PostDestinationFactoryManager
from nonebot_plugin_pixivbot.protocol_dep.postman import PostmanManager
from nonebot_plugin_pixivbot.service.pixiv_account_binder import PixivAccountBinder
from nonebot_plugin_pixivbot.utils.lifecycler import on_bot_connect, on_bot_disconnect
from nonebot_plugin_pixivbot.utils.nonebot import get_adapter_name

UID = TypeVar("UID")
GID = TypeVar("GID")

PD = PostDestination[UID, GID]
ID = PostIdentifier[UID, GID]


@context.inject
@context.root.register_eager_singleton()
class Watchman:
    conf = Inject(Config)
    apscheduler = Inject(AsyncIOScheduler)
    repo = Inject(WatchTaskRepo)
    binder = Inject(PixivAccountBinder)
    postman_mgr = Inject(PostmanManager)
    pd_factory_mgr = Inject(PostDestinationFactoryManager)
    auth_mgr = Inject(AuthenticatorManager)

    def __init__(self):
        on_bot_connect(self.on_bot_connect, replay=True)
        on_bot_disconnect(self.on_bot_disconnect)

    @staticmethod
    def _make_job_id(task: WatchTask[UID, GID]):
        return f'watchman {task.subscriber} {task.code}'

    _trigger_hasher_mapper = {
        WatchType.user_illusts: lambda args: args["user_id"],
        WatchType.following_illusts: lambda args: hash(args["sender_user_id"]),
    }

    _handlers = {
        WatchType.user_illusts: context.require(WatchUserIllustsHandler),
        WatchType.following_illusts: context.require(WatchFollowingIllustsHandler),
    }

    async def _on_trigger(self, task: WatchTask[UID, GID], post_dest: PD):
        logger.info(f"[watchman] triggered {task}")

        try:
            try:
                await self._handlers[task.type].handle_with_parsed_args(post_dest=post_dest, silently=True, task=task)
            except Exception as e:
                # 保存checkpoint，避免一次异常后下一次重复推送
                # 但是会存在丢失推送的问题
                task.checkpoint = datetime.now(timezone.utc)
                await self.repo.update(task)

                raise e
        except ActionFailed as e:
            logger.error("[watchman] ActionFailed" + str(e))

            available = self.auth_mgr.available(post_dest)
            if isawaitable(available):
                available = await available

            if not available:
                logger.info(f"[watchman] {post_dest} is no longer available, removing all his tasks...")
                await self.unwatch_all_by_subscriber(post_dest.identifier)

    def _make_job_trigger(self, task: WatchTask):
        hasher = self._trigger_hasher_mapper[task.type]
        hash_sec = hasher(task.kwargs) % self.conf.pixiv_watch_interval
        yesterday = datetime.now(timezone.utc).replace(
            hour=0, minute=0, second=0, microsecond=0
        ) + timedelta(days=-1) + timedelta(seconds=hash_sec)
        trigger = IntervalTrigger(seconds=self.conf.pixiv_watch_interval, start_date=yesterday)
        return trigger

    def _add_job(self, task: WatchTask[UID, GID], post_dest: PD):
        job_id = self._make_job_id(task)
        trigger = self._make_job_trigger(task)
        self.apscheduler.add_job(self._on_trigger, id=job_id, trigger=trigger,
                                 args=[task, post_dest],
                                 max_instances=1)
        logger.success(f"[watchman] add job \"{job_id}\" on {trigger}")

    def _remove_job(self, task: WatchTask[UID, GID]):
        job_id = self._make_job_id(task)
        self.apscheduler.remove_job(job_id)
        logger.success(f"[watchman] remove job \"{job_id}\"")

    async def on_bot_connect(self, bot: Bot):
        adapter = get_adapter_name(bot)
        async for task in self.repo.get_by_adapter(adapter):
            pd = self.pd_factory_mgr.build(bot, task.subscriber.user_id, task.subscriber.group_id)
            self._add_job(task, pd)

    async def on_bot_disconnect(self, bot: Bot):
        adapter = get_adapter_name(bot)
        async for task in self.repo.get_by_adapter(adapter):
            try:
                self._remove_job(task)
            except Exception as e:
                logger.error(f"[watchman] error occurred when remove job "
                             f"\"{self._make_job_id(task)}\"")
                logger.exception(e)

    async def watch(self, type_: WatchType,
                    kwargs: Dict[str, Any],
                    subscriber: PD) -> bool:
        task = WatchTask(type=type_, kwargs=kwargs, subscriber=subscriber.identifier)
        ok = await self.repo.insert(task)
        if ok:
            logger.success(f"[watchman] successfully inserted subscription {task}")
            self._add_job(task, subscriber.normalized())
        return ok

    async def unwatch(self, type_: WatchType, code: int) -> bool:
        task = await self.repo.delete_one(type_, code)
        if task:
            logger.success(f"[scheduler] successfully removed subscription {task}")
            self._remove_job(task)
            return True
        else:
            return False

    async def unwatch_all_by_subscriber(self, subscriber: ID):
        old = await self.repo.delete_many_by_subscriber(subscriber)
        for task in old:
            logger.success(f"[scheduler] successfully removed subscription {task}")
            self._remove_job(task)

    def get_by_subscriber(self, subscriber: ID) -> AsyncIterable[WatchTask]:
        return self.repo.get_by_subscriber(subscriber)



from enum import Enum
from typing import TypeVar, Generic, Sequence, Dict, Any

import tzlocal
from pydantic.generics import GenericModel

from nonebot_plugin_pixivbot.model import PostIdentifier

UID = TypeVar("UID")
GID = TypeVar("GID")


class ScheduleType(str, Enum):
    random_bookmark = "random_bookmark"
    random_recommended_illust = "random_recommended_illust"
    random_illust = "random_illust"
    random_user_illust = "random_user_illust"
    ranking = "ranking"


class Subscription(GenericModel, Generic[UID, GID]):
    code: int = 0
    type: ScheduleType
    kwargs: Dict[str, Any]
    subscriber: PostIdentifier[UID, GID]
    schedule: Sequence[int]
    tz: str = tzlocal.get_localzone_name()

    class Config:
        orm_mode = True


__all__ = ("Subscription", "ScheduleType")



from typing import TypeVar, Sequence, Dict, Any

from nonebot_plugin_pixivbot.context import Inject
from nonebot_plugin_pixivbot.handler.interceptor.permission_interceptor import GroupAdminInterceptor, \
    AnyPermissionInterceptor, SuperuserInterceptor
from nonebot_plugin_pixivbot.model import PostIdentifier, WatchType
from nonebot_plugin_pixivbot.protocol_dep.post_dest import PostDestination
from nonebot_plugin_pixivbot.service.pixiv_service import PixivService
from nonebot_plugin_pixivbot.service.watchman import Watchman
from nonebot_plugin_pixivbot.utils.errors import BadRequestError
from .command import CommandHandler, SubCommandHandler
from ..pkg_context import context

UID = TypeVar("UID")
GID = TypeVar("GID")


async def parse_and_get_user(raw_user: str):
    pixiv = context.require(PixivService)
    # try parse
    try:
        user = int(raw_user)
        return await pixiv.get_user(user)
    except ValueError:
        return await pixiv.get_user(raw_user)


async def build_tasks_msg(identifier: PostIdentifier):
    watchman = context.require(Watchman)
    tasks = [x async for x in watchman.get_by_subscriber(identifier)]
    msg = "当前订阅：\n"
    if len(tasks) > 0:
        for x in tasks:
            args = list(filter(lambda kv: kv[1], x.kwargs.items()))
            if len(args) != 0:
                args_text = " ".join(map(lambda kv: f'{kv[0]}={kv[1]}', args))
                args_text = f"({args_text})"
            else:
                args_text = ""
            msg += f'[{x.code}] {x.type.name} {args_text}\n'
    else:
        msg += '无\n'
    return msg


async def parse_user_illusts_args(args: Sequence[str]):
    if len(args) < 2:
        raise BadRequestError()

    user = await parse_and_get_user(args[1])

    watch_args = {"user_id": user.id}
    message = f"{user.name}({user.id})老师的插画更新"

    return watch_args, message


async def parse_following_illusts_args(args: Sequence[str], post_dest: PostDestination[UID, GID]):
    if len(args) > 1:
        user = await parse_and_get_user(args[1])

        watch_args = {"pixiv_user_id": user.id,
                      "sender_user_id": post_dest.user_id}
        message = f"{user.name}({user.id})用户的关注者插画更新"
    else:
        watch_args = {"pixiv_user_id": 0,
                      "sender_user_id": post_dest.user_id}
        message = f"关注者插画更新"

    return watch_args, message


@context.inject
@context.require(CommandHandler).sub_command("watch")
class WatchHandler(SubCommandHandler):
    watchman = Inject(Watchman)

    def __init__(self):
        super().__init__()
        self.add_interceptor(AnyPermissionInterceptor(
            context.require(SuperuserInterceptor),
            context.require(GroupAdminInterceptor)
        ))

    @classmethod
    def type(cls) -> str:
        return "watch"

    def enabled(self) -> bool:
        return True

    async def parse_args(self, args: Sequence[str], post_dest: PostDestination[UID, GID]) -> dict:
        if len(args) == 0:
            raise BadRequestError()

        try:
            type = WatchType[args[0]]
            if type == WatchType.user_illusts:
                watch_kwargs, message = await parse_user_illusts_args(args)
            elif type == WatchType.following_illusts:
                watch_kwargs, message = await parse_following_illusts_args(args, post_dest)
            else:
                raise KeyError()
        except KeyError as e:
            raise BadRequestError(f"未知订阅类型：{args[0]}") from e

        return {"type": type, "watch_kwargs": watch_kwargs, "success_message": "成功订阅" + message}

    # noinspection PyMethodOverriding
    async def actual_handle(self, *, type: WatchType,
                            watch_kwargs: Dict[str, Any],
                            success_message: str,
                            post_dest: PostDestination[UID, GID],
                            silently: bool = False, **kwargs):
        ok = await self.watchman.watch(type, watch_kwargs, post_dest)
        if ok:
            await self.post_plain_text(success_message, post_dest)
        else:
            await self.post_plain_text("该订阅已存在", post_dest)

    async def actual_handle_bad_request(self, err: BadRequestError,
                                        *, post_dest: PostDestination[UID, GID],
                                        silently: bool = False):
        msg = ""
        if err.message:
            msg += err.message
            msg += '\n\n'

        msg += await build_tasks_msg(post_dest.identifier)
        msg += "\n" \
               "命令格式：/pixivbot watch <type> [..args]\n" \
               "参数：\n" \
               "  <type>：可选值有user_illusts, following_illusts\n" \
               "  [...args]：根据<type>不同需要提供不同的参数\n" \
               "示例：/pixivbot watch user_illusts <用户名>\n"
        await self.post_plain_text(message=msg, post_dest=post_dest)


@context.inject
@context.require(CommandHandler).sub_command("unwatch")
class UnwatchHandler(SubCommandHandler):
    watchman = Inject(Watchman)

    def __init__(self):
        super().__init__()
        self.add_interceptor(AnyPermissionInterceptor(
            context.require(SuperuserInterceptor),
            context.require(GroupAdminInterceptor)
        ))

    @classmethod
    def type(cls) -> str:
        return "unwatch"

    def enabled(self) -> bool:
        return True

    def parse_args(self, args: Sequence[str], post_dest: PostDestination[UID, GID]) -> dict:
        if len(args) == 0:
            raise BadRequestError()
        try:
            return {"code": int(args[0])}
        except ValueError as e:
            raise BadRequestError(f"不合法的订阅编号：{args[0]}") from e

    # noinspection PyMethodOverriding
    async def actual_handle(self, *, code: int,
                            post_dest: PostDestination[UID, GID],
                            silently: bool = False, **kwargs):
        if await self.watchman.unwatch(post_dest.identifier, code):
            await self.post_plain_text(message="取消订阅成功", post_dest=post_dest)
        else:
            raise BadRequestError("取消订阅失败，不存在该订阅")

    async def actual_handle_bad_request(self, err: BadRequestError,
                                        *, post_dest: PostDestination[UID, GID],
                                        silently: bool = False):
        msg = ""
        if err.message:
            msg += err.message
            msg += '\n'

        msg += await build_tasks_msg(post_dest.identifier)
        msg += "\n"
        msg += "命令格式：/pixivbot unwatch <id>"
        await self.post_plain_text(message=msg, post_dest=post_dest)



from typing import TypeVar, Sequence

from nonebot_plugin_pixivbot.context import Inject
from nonebot_plugin_pixivbot.handler.interceptor.permission_interceptor import GroupAdminInterceptor, \
    AnyPermissionInterceptor, SuperuserInterceptor
from nonebot_plugin_pixivbot.model import PostIdentifier, ScheduleType
from nonebot_plugin_pixivbot.protocol_dep.post_dest import PostDestination
from nonebot_plugin_pixivbot.service.scheduler import Scheduler
from nonebot_plugin_pixivbot.utils.errors import BadRequestError
from .command import CommandHandler, SubCommandHandler
from ..pkg_context import context

UID = TypeVar("UID")
GID = TypeVar("GID")


async def build_subscriptions_msg(subscriber: PostIdentifier[UID, GID]):
    scheduler = context.require(Scheduler)
    subscription = [x async for x in scheduler.get_by_subscriber(subscriber)]
    msg = "当前订阅：\n"
    if len(subscription) > 0:
        for x in subscription:
            args = list(filter(lambda kv: kv[1], x.kwargs.items()))
            if len(args) != 0:
                args_text = " ".join(map(lambda kv: f'{kv[0]}={kv[1]}', args))
                args_text = f"({args_text})"
            else:
                args_text = ""
            schedule_text = f'{str(x.schedule[0]).zfill(2)}:{str(x.schedule[1]).zfill(2)}' \
                            f'+{str(x.schedule[2]).zfill(2)}:{str(x.schedule[3]).zfill(2)}*x'
            msg += f'[{x.code}] {x.type.name} {schedule_text} {args_text}\n'
    else:
        msg += '无\n'
    return msg


@context.inject
@context.require(CommandHandler).sub_command("schedule")
class ScheduleHandler(SubCommandHandler):
    scheduler = Inject(Scheduler)

    def __init__(self):
        super().__init__()
        self.add_interceptor(AnyPermissionInterceptor(
            context.require(SuperuserInterceptor),
            context.require(GroupAdminInterceptor)
        ))

    @classmethod
    def type(cls) -> str:
        return "schedule"

    def enabled(self) -> bool:
        return True

    def parse_args(self, args: Sequence[str], post_dest: PostIdentifier[UID, GID]) -> dict:
        if len(args) < 2:
            raise BadRequestError()
        try:
            return {"type": ScheduleType(args[0]),
                    "schedule": args[1],
                    "args": args[2:]}
        except ValueError as e:
            raise BadRequestError(f"未知订阅类型：{args[0]}") from e

    # noinspection PyMethodOverriding
    async def actual_handle(self, *, type: ScheduleType,
                            schedule: str,
                            args: Sequence[str],
                            post_dest: PostDestination[UID, GID],
                            silently: bool = False, **kwargs):
        await self.scheduler.schedule(type, schedule, args, post_dest)
        await self.post_plain_text(message="订阅成功", post_dest=post_dest)

    async def actual_handle_bad_request(self, err: BadRequestError,
                                        *, post_dest: PostDestination[UID, GID],
                                        silently: bool = False):
        msg = ""
        if err.message:
            msg += err.message
            msg += '\n'

        msg += await build_subscriptions_msg(post_dest.identifier)
        msg += "\n" \
               "命令格式：/pixivbot schedule <type> <schedule> [..args]\n" \
               "参数：\n" \
               "  <type>：可选值有random_bookmark, random_recommended_illust, random_illust, " \
               "random_user_illust, ranking\n" \
               "  <schedule>：格式为HH:mm（每日固定时间点推送）或HH:mm*x（间隔时间推送）\n" \
               "  [...args]：根据<type>不同需要提供不同的参数\n" \
               "示例：/pixivbot schedule ranking 06:00*x day 1-5"
        await self.post_plain_text(message=msg, post_dest=post_dest)


@context.inject
@context.require(CommandHandler).sub_command("unschedule")
class UnscheduleHandler(SubCommandHandler):
    scheduler = Inject(Scheduler)

    def __init__(self):
        super().__init__()
        self.add_interceptor(AnyPermissionInterceptor(
            context.require(SuperuserInterceptor),
            context.require(GroupAdminInterceptor)
        ))

    @classmethod
    def type(cls) -> str:
        return "unschedule"

    def enabled(self) -> bool:
        return True

    def parse_args(self, args: Sequence[str], post_dest: PostIdentifier[UID, GID]) -> dict:
        if len(args) == 0:
            raise BadRequestError()
        try:
            return {"code": int(args[0])}
        except ValueError as e:
            raise BadRequestError(f"不合法的订阅编号：{args[0]}") from e

    # noinspection PyMethodOverriding
    async def actual_handle(self, *, code: int,
                            post_dest: PostDestination[UID, GID],
                            silently: bool = False):
        if await self.scheduler.unschedule(post_dest.identifier, code):
            await self.post_plain_text(message="取消订阅成功", post_dest=post_dest)
        else:
            raise BadRequestError("取消订阅失败，不存在该订阅")

    async def actual_handle_bad_request(self, err: BadRequestError,
                                        *, post_dest: PostDestination[UID, GID],
                                        silently: bool = False):
        msg = ""
        if err.message:
            msg += err.message
            msg += '\n'

        msg += await build_subscriptions_msg(post_dest.identifier)
        msg += "\n"
        msg += "命令格式：/pixivbot unschedule <id>"
        await self.post_plain_text(message=msg, post_dest=post_dest)



from motor.motor_asyncio import AsyncIOMotorDatabase

from nonebot_plugin_pixivbot.data.source.mongo.migration.mongo_migration import MongoMigration
from nonebot_plugin_pixivbot.data.source.mongo.migration.mongo_migration_manager import MongoMigrationManager
from nonebot_plugin_pixivbot.global_context import context


@context.require(MongoMigrationManager).register
class MongoV1ToV2(MongoMigration):
    from_db_version = 1
    to_db_version = 2

    async def migrate(self, db: AsyncIOMotorDatabase):
        await self.migrate_pixiv_binding(db)
        await self.migrate_subscription(db)

    async def migrate_pixiv_binding(self, db: AsyncIOMotorDatabase):
        try:
            await db["pixiv_binding"].drop_index("qq_id_1")
        except:
            pass

        # 先drop index，否则报duplicate key

        await db["pixiv_binding"].update_many(
            {},
            [
                {"$set": {"adapter": "onebot", "user_id": "$qq_id"}},
                {"$unset": ["qq_id"]}
            ]
        )

    async def migrate_subscription(self, db: AsyncIOMotorDatabase):
        try:
            await db["subscription"].drop_index("user_id_1")
        except:
            pass

        try:
            await db["subscription"].drop_index("group_id_1")
        except:
            pass

        try:
            await db["subscription"].drop_index("type_1_user_id_1")
        except:
            pass

        try:
            await db["subscription"].drop_index("type_1_group_id_1")
        except:
            pass

        await db["subscription"].update_many(
            {},
            {"$set": {"adapter": "onebot"}}
        )


__all__ = ("MongoV1ToV2",)



from abc import ABC, abstractmethod

from motor.motor_asyncio import AsyncIOMotorDatabase


class MongoMigration(ABC):
    from_db_version: int
    to_db_version: int

    @abstractmethod
    async def migrate(self, client: AsyncIOMotorDatabase):
        raise NotImplementedError()



from __future__ import annotations

import re
from datetime import datetime, timedelta
from inspect import isawaitable
from typing import TypeVar, Dict, Sequence, Union, TYPE_CHECKING, AsyncIterable

import pytz
from apscheduler.schedulers.asyncio import AsyncIOScheduler
from apscheduler.triggers.interval import IntervalTrigger
from lazy import lazy
from nonebot import logger, Bot
from nonebot.exception import ActionFailed

from nonebot_plugin_pixivbot.context import Inject
from nonebot_plugin_pixivbot.data.subscription import SubscriptionRepo
from nonebot_plugin_pixivbot.global_context import context
from nonebot_plugin_pixivbot.model import Subscription, PostIdentifier
from nonebot_plugin_pixivbot.model.subscription import ScheduleType
from nonebot_plugin_pixivbot.protocol_dep.authenticator import AuthenticatorManager
from nonebot_plugin_pixivbot.protocol_dep.post_dest import PostDestination, PostDestinationFactoryManager
from nonebot_plugin_pixivbot.utils.errors import BadRequestError
from nonebot_plugin_pixivbot.utils.lifecycler import on_bot_connect, on_bot_disconnect
from nonebot_plugin_pixivbot.utils.nonebot import get_adapter_name

if TYPE_CHECKING:
    from nonebot_plugin_pixivbot.handler import Handler

UID = TypeVar("UID")
GID = TypeVar("GID")
ID = PostIdentifier[UID, GID]


def parse_schedule(raw_schedule: str) -> Sequence[int]:
    start_only_mat = re.fullmatch(r'(\d+):(\d+)', raw_schedule)
    if start_only_mat is not None:
        g = start_only_mat.groups()
        start_hour, start_minute = int(g[0]), int(g[1])
        interval_hour, interval_minute = 24, 0
    else:
        interval_only_mat = re.fullmatch(
            r'(\d+):(\d+)\*x', raw_schedule)
        if interval_only_mat is not None:
            g = interval_only_mat.groups()
            start_hour, start_minute = 0, 0
            interval_hour, interval_minute = int(g[0]), int(g[1])
        else:
            mat = re.fullmatch(
                r'(\d+):(\d+)\+(\d+):(\d+)\*x', raw_schedule)
            if mat is not None:
                g = mat.groups()
                start_hour, start_minute = int(g[0]), int(g[1])
                interval_hour, interval_minute = int(g[2]), int(g[3])
            else:
                raise BadRequestError(f'{raw_schedule}不是合法的时间')

    if start_hour < 0 or start_hour >= 24 or start_minute < 0 or start_minute >= 60 \
            or interval_hour < 0 or interval_minute >= 60 \
            or (interval_hour > 0 and interval_minute < 0) or (interval_hour == 0 and interval_minute <= 0):
        raise BadRequestError(f'{raw_schedule}不是合法的时间')

    return start_hour, start_minute, interval_hour, interval_minute


@context.inject
@context.register_eager_singleton()
class Scheduler:
    apscheduler: AsyncIOScheduler = Inject(AsyncIOScheduler)
    repo: SubscriptionRepo = Inject(SubscriptionRepo)
    pd_factory_mgr: PostDestinationFactoryManager = Inject(PostDestinationFactoryManager)
    auth_mgr: AuthenticatorManager = Inject(AuthenticatorManager)

    def __init__(self):
        on_bot_connect(self.on_bot_connect, replay=True)
        on_bot_disconnect(self.on_bot_disconnect)

    @staticmethod
    def _make_job_id(sub: Subscription):
        return f'scheduler {sub.subscriber} {sub.code}'

    @lazy
    def _handlers(self) -> Dict[ScheduleType, Handler]:
        # 解决Handler和Scheduler的循环引用
        from nonebot_plugin_pixivbot.handler.common import RandomBookmarkHandler, RandomRecommendedIllustHandler, \
            RankingHandler, RandomIllustHandler, RandomUserIllustHandler
        return {
            ScheduleType.random_bookmark: context.require(RandomBookmarkHandler),
            ScheduleType.random_recommended_illust: context.require(RandomRecommendedIllustHandler),
            ScheduleType.ranking: context.require(RankingHandler),
            ScheduleType.random_illust: context.require(RandomIllustHandler),
            ScheduleType.random_user_illust: context.require(RandomUserIllustHandler),
        }

    async def _on_trigger(self, sub: Subscription[UID, GID], post_dest: PostDestination[UID, GID]):
        logger.info(f"[scheduler] triggered {sub}")

        try:
            await self._handlers[sub.type].handle_with_parsed_args(post_dest=post_dest, silently=True, **sub.kwargs)
        except ActionFailed as e:
            logger.warning("[scheduler] ActionFailed: " + str(e))

            available = self.auth_mgr.available(post_dest)
            if isawaitable(available):
                available = await available

            if not available:
                logger.info(f"[scheduler] {post_dest} is no longer available, removing all his subscriptions...")
                await self.unschedule_all_by_subscriber(post_dest.identifier)

    def _add_job(self, sub: Subscription[UID, GID], post_dest: PostDestination[UID, GID]):
        offset_hour, offset_minute, hours, minutes = sub.schedule
        tz = pytz.timezone(sub.tz)
        trigger = IntervalTrigger(hours=hours, minutes=minutes,
                                  start_date=datetime.now(tz).replace(hour=offset_hour,
                                                                      minute=offset_minute,
                                                                      second=0,
                                                                      microsecond=0) + timedelta(days=-1))

        job_id = self._make_job_id(sub)
        self.apscheduler.add_job(self._on_trigger, id=job_id, trigger=trigger,
                                 kwargs={"sub": sub, "post_dest": post_dest})
        logger.success(f"[scheduler] added job \"{job_id}\" on {trigger}")

    def _remove_job(self, sub: Subscription[UID, GID]):
        job_id = self._make_job_id(sub)
        self.apscheduler.remove_job(job_id)
        logger.success(f"[scheduler] removed job \"{job_id}\"")

    async def on_bot_connect(self, bot: Bot):
        adapter = get_adapter_name(bot)
        async for sub in self.repo.get_by_adapter(adapter):
            post_dest = self.pd_factory_mgr.build(bot, sub.subscriber.user_id, sub.subscriber.group_id)
            self._add_job(sub, post_dest)

    async def on_bot_disconnect(self, bot: Bot):
        adapter = get_adapter_name(bot)
        async for sub in self.repo.get_by_adapter(adapter):
            try:
                self._remove_job(sub)
            except Exception as e:
                logger.error(f"[scheduler] error occurred when remove job "
                             f"\"{self._make_job_id(sub)}\"")
                logger.exception(e)

    async def schedule(self, type_: ScheduleType,
                       schedule: Union[str, Sequence[int]],
                       args: Sequence[str],
                       post_dest: PostDestination[UID, GID]):
        if args is None:
            args = []

        if isinstance(schedule, str):
            schedule = parse_schedule(schedule)

        kwargs = self._handlers[type_].parse_args(args, post_dest)
        if isawaitable(kwargs):
            kwargs = await kwargs

        sub = Subscription(type=type_, kwargs=kwargs, subscriber=post_dest.identifier, schedule=schedule)
        await self.repo.insert(sub)
        logger.success(f"[scheduler] successfully inserted subscription {sub}")
        self._add_job(sub, post_dest.normalized())

    async def unschedule(self, subscriber: PostIdentifier[UID, GID], code: int) -> bool:
        sub = await self.repo.delete_one(subscriber, code)
        if sub:
            logger.success(f"[scheduler] successfully removed subscription {sub}")
            self._remove_job(sub)
            return True
        else:
            return False

    async def unschedule_all_by_subscriber(self, subscriber: PostIdentifier[UID, GID]):
        old = await self.repo.delete_many_by_subscriber(subscriber)
        for sub in old:
            logger.success(f"[scheduler] successfully removed subscription {sub}")
            self._remove_job(sub)

    def get_by_subscriber(self, subscriber: PostIdentifier[UID, GID]) -> AsyncIterable[Subscription]:
        return self.repo.get_by_subscriber(subscriber)


__all__ = ("Scheduler",)



from typing import TypeVar

from nonebot_plugin_pixivbot.model import PostIdentifier

UID = TypeVar("UID")
GID = TypeVar("GID")

ID = PostIdentifier[UID, GID]


def process_subscriber(subscriber: ID) -> PostIdentifier[UID, GID]:
    if subscriber.group_id:
        return PostIdentifier(subscriber.adapter, None, subscriber.group_id)
    elif subscriber.user_id:
        return PostIdentifier(subscriber.adapter, subscriber.user_id, None)
    else:
        raise ValueError("at least one of user_id and group_id should be not None")



from typing import TypeVar, Optional, Protocol, AsyncIterable, Collection

from nonebot_plugin_pixivbot import context
from nonebot_plugin_pixivbot.config import Config
from nonebot_plugin_pixivbot.model import PostIdentifier, WatchTask
from ...enums import DataSourceType

UID = TypeVar("UID")
GID = TypeVar("GID")

ID = PostIdentifier[UID, GID]


class WatchTaskRepo(Protocol):
    def get_by_subscriber(self, subscriber: ID) -> AsyncIterable[WatchTask]:
        ...

    def get_by_adapter(self, adapter: str) -> AsyncIterable[WatchTask]:
        ...

    async def get_by_code(self, subscriber: ID, code: int) -> Optional[WatchTask]:
        ...

    async def insert(self, task: WatchTask) -> bool:
        ...

    async def update(self, task: WatchTask) -> bool:
        ...

    async def delete_one(self, subscriber: ID, code: int) -> Optional[WatchTask]:
        ...

    async def delete_many_by_subscriber(self, subscriber: ID) -> Collection[WatchTask]:
        ...


conf = context.require(Config)
if conf.pixiv_data_source == DataSourceType.mongo:
    from .mongo import MongoWatchTaskRepo

    context.bind(WatchTaskRepo, MongoWatchTaskRepo)
else:
    from .sql import SqlWatchTaskRepo

    context.bind(WatchTaskRepo, SqlWatchTaskRepo)

__all__ = ("WatchTaskRepo",)



from abc import ABC, abstractmethod
from typing import Dict, TypeVar, Generic, Type

from nonebot_plugin_pixivbot.global_context import context


class ProtocolDep(ABC):
    @classmethod
    @abstractmethod
    def adapter(cls) -> str:
        raise NotImplementedError()


T = TypeVar("T", bound=ProtocolDep)


class ProtocolDepManager(Generic[T], ABC):
    def __init__(self):
        self.factories: Dict[str, Type[T]] = {}

    def register(self, cls: Type[T]):
        self.factories[cls.adapter()] = cls
        if cls not in context:
            context.register_singleton()(cls)
        return cls

    def require(self, adapter: str):
        t = self.factories.get(adapter, None)
        if t is None:
            raise RuntimeError(f"未找到{adapter}协议的特化插件，请尝试安装nonebot-plugin-pixivbot-{adapter}")
        return context.require(t)

    def __getitem__(self, adapter: str):
        return self.require(adapter)


__all__ = ("ProtocolDep", "ProtocolDepManager")



from io import BytesIO
from typing import Optional

from pydantic import BaseModel
from tzlocal import get_localzone

from nonebot_plugin_pixivbot.config import Config
from nonebot_plugin_pixivbot.data.pixiv_repo import PixivRepo
from nonebot_plugin_pixivbot.enums import BlockAction
from nonebot_plugin_pixivbot.global_context import context
from nonebot_plugin_pixivbot.model import Illust

conf = context.require(Config)


async def download_image(illust: Illust):
    with BytesIO() as bio:
        repo = context.require(PixivRepo)
        async for x in repo.image(illust):
            bio.write(x)
            break
        return bio.getvalue()


class IllustMessageModel(BaseModel):
    id: int = 0
    title: str = ""
    author: str = ""
    create_time: str = ""
    link: str = ""
    image: bytes = bytes(0)

    header: Optional[str] = None
    number: Optional[int] = None

    block_action: Optional[BlockAction] = None
    block_message: str = ""

    @staticmethod
    async def from_illust(illust: Illust, *,
                          header: Optional[str] = None,
                          number: Optional[int] = None) -> Optional["IllustMessageModel"]:
        model = IllustMessageModel(id=illust.id, header=header, number=number)

        if illust.has_tags(conf.pixiv_block_tags):
            model.block_action = conf.pixiv_block_action
            if conf.pixiv_block_action == BlockAction.no_image:
                model.block_message = "该画像因含有不可描述的tag而被自主规制"
            elif conf.pixiv_block_action == BlockAction.completely_block:
                model.block_message = "该画像因含有不可描述的tag而被自主规制"
                return model
            elif conf.pixiv_block_action == BlockAction.no_reply:
                return None
        else:
            model.image = await download_image(illust)
        model.title = illust.title
        model.author = f"{illust.user.name} ({illust.user.id})"
        model.create_time = illust.create_date.astimezone(get_localzone()).strftime('%Y-%m-%d %H:%M:%S')
        model.link = f"https://www.pixiv.net/artworks/{illust.id}"

        return model



# from typing import TypeVar, Generic, Sequence, Dict, Any, Optional
#
# from pydantic import validator
# from pydantic.generics import GenericModel
#
# from nonebot_plugin_pixivbot.postman import PostIdentifier
#
# UID = TypeVar("UID")
# GID = TypeVar("GID")
#
#
# class SubscriptionV1(GenericModel, Generic[UID, GID]):
#     user_id: Optional[UID]
#     group_id: Optional[GID]
#     type: str
#     kwargs: Dict[str, Any]
#     schedule: Sequence[int]
#
#     @validator("user_id")
#     def validate(cls, user_id, values, **kwargs):
#         group_id = None
#         if "group_id" in values:
#             group_id = values["group_id"]
#
#         if not user_id and not group_id:
#             raise ValueError("at least one of user_id and group_id should be not None")
#         return user_id
#
#     @property
#     def identifier(self):
#         return PostIdentifier("onebot", self.user_id, self.group_id)



from nonebot import Bot
from nonebot.internal.adapter import Event
from nonebot.rule import to_me
from nonebot.typing import T_State

from nonebot_plugin_pixivbot.config import Config
from nonebot_plugin_pixivbot.global_context import context
from nonebot_plugin_pixivbot.protocol_dep.post_dest import PostDestinationFactoryManager
from nonebot_plugin_pixivbot.utils.decode_integer import decode_integer
from nonebot_plugin_pixivbot.utils.errors import BadRequestError


def get_count(state: T_State, pos: int = 0):
    count = 1
    if "_matched_groups" in state:
        raw_count = state["_matched_groups"][pos]
        if raw_count:
            try:
                count = decode_integer(raw_count)
            except:
                raise BadRequestError(f"{raw_count}不是合法的数字")
    return count


def get_common_query_rule():
    if context.require(Config).pixiv_query_to_me_only:
        rule = to_me()
    else:
        rule = None
    return rule


def get_command_rule():
    if context.require(Config).pixiv_command_to_me_only:
        rule = to_me()
    else:
        rule = None
    return rule


def get_post_dest(bot: Bot, event: Event):
    return context.require(PostDestinationFactoryManager).from_event(bot, event)


__all__ = ("get_count", "get_common_query_rule", "get_command_rule", "get_post_dest")



class LazyDelegation:
    __slots__ = ("__cache__", "__builder__")

    def __init__(self, builder):
        object.__setattr__(self, "__cache__", None)
        object.__setattr__(self, "__builder__", builder)

    def __getattribute__(self, item):
        if object.__getattribute__(self, "__cache__") is None:
            object.__setattr__(self, "__cache__", object.__getattribute__(self, "__builder__")())
        return getattr(object.__getattribute__(self, "__cache__"), item)



from typing import Type, Optional

from motor.motor_asyncio import AsyncIOMotorDatabase
from nonebot import logger

from nonebot_plugin_pixivbot.data.source.mongo.migration.mongo_migration import MongoMigration
from nonebot_plugin_pixivbot.global_context import context


@context.register_singleton()
class MongoMigrationManager:
    def __init__(self):
        self._mapping = dict[int, list[MongoMigration]]()

    def add(self, migration: MongoMigration):
        if migration.from_db_version not in self._mapping:
            self._mapping[migration.from_db_version] = []
        self._mapping[migration.from_db_version].append(migration)

    def register(self, migration_cls: Type[MongoMigration]) -> Type[MongoMigration]:
        self.add(migration_cls())
        return migration_cls

    async def perform_migration(self, client: AsyncIOMotorDatabase, from_db_version: int, to_db_version: int):
        while from_db_version < to_db_version:
            from_migrations = self._mapping.get(from_db_version, [])

            choice: Optional[MongoMigration] = None
            for mig in from_migrations:
                if choice is None or (choice.to_db_version < mig.to_db_version <= to_db_version):
                    choice = mig

            if choice is None:
                raise NoMigrationError(from_db_version, to_db_version)

            await choice.migrate(client)
            logger.success(f"migrated from {from_db_version} to {choice.to_db_version}")
            from_db_version = choice.to_db_version


class NoMigrationError(Exception):
    def __init__(self, from_db_version: int, to_db_version: int):
        super().__init__()
        self.from_db_version = from_db_version
        self.to_db_version = to_db_version

    def __str__(self):
        return f"no migration from {self.from_db_version} to {self.to_db_version}"



from datetime import datetime, timedelta
from typing import Any, AsyncGenerator, Union

from frozendict import frozendict
from nonebot import logger
from pydantic import BaseModel

from nonebot_plugin_pixivbot import context
from nonebot_plugin_pixivbot.config import Config
from nonebot_plugin_pixivbot.context import Inject
from nonebot_plugin_pixivbot.enums import RankingMode
from nonebot_plugin_pixivbot.model import Illust, User, UserPreview
from nonebot_plugin_pixivbot.utils.shared_agen import SharedAsyncGeneratorManager
from .enums import PixivResType, CacheStrategy
from .lazy_illust import LazyIllust
from .local_repo import LocalPixivRepo
from .mediator import mediate_single, mediate_many, mediate_append
from .models import PixivRepoMetadata
from .remote_repo import RemotePixivRepo


class SharedAgenIdentifier(BaseModel):
    type: PixivResType
    kwargs: frozendict[str, Any]

    def __init__(self, type: PixivResType, **kwargs):
        super().__init__(type=type, kwargs=frozendict(kwargs))

    def __str__(self):
        return f"({self.type.name} {', '.join(map(lambda k: f'{k}={self.kwargs[k]}', {**self.kwargs}))})"

    class Config:
        frozen = True


@context.inject
@context.register_singleton()
class PixivSharedAsyncGeneratorManager(SharedAsyncGeneratorManager[SharedAgenIdentifier, Any]):
    log_tag = "pixiv_shared_agen"

    conf = Inject(Config)
    local = Inject(LocalPixivRepo)
    remote = Inject(RemotePixivRepo)

    def illust_detail_factory(self, illust_id: int,
                              cache_strategy: CacheStrategy) -> AsyncGenerator[Illust, None]:
        return mediate_single(
            cache_factory=self.local.illust_detail,
            remote_factory=self.remote.illust_detail,
            query_kwargs={"illust_id": illust_id},
            cache_updater=lambda data, metadata: self.local.update_illust_detail(data, metadata),
            force_expiration=cache_strategy == CacheStrategy.FORCE_EXPIRATION,
        )

    def user_detail_factory(self, user_id: int,
                            cache_strategy: CacheStrategy) -> AsyncGenerator[User, None]:
        return mediate_single(
            cache_factory=self.local.user_detail,
            remote_factory=self.remote.user_detail,
            query_kwargs={"user_id": user_id},
            cache_updater=lambda data, metadata: self.local.update_user_detail(data, metadata),
            force_expiration=cache_strategy == CacheStrategy.FORCE_EXPIRATION,
        )

    def search_illust_factory(self, word: str,
                              cache_strategy: CacheStrategy) -> AsyncGenerator[LazyIllust, None]:
        return mediate_append(
            cache_factory=self.local.search_illust,
            remote_factory=self.remote.search_illust,
            query_kwargs={"word": word},
            cache_appender=lambda data, metadata: self.local.append_search_illust(word, data, metadata),
            max_item=self.conf.pixiv_random_illust_max_item,
            max_page=self.conf.pixiv_random_illust_max_page,
            force_expiration=cache_strategy == CacheStrategy.FORCE_EXPIRATION,
        )

    def search_user_factory(self, word: str,
                            cache_strategy: CacheStrategy) -> AsyncGenerator[User, None]:
        return mediate_append(
            cache_factory=self.local.search_user,
            remote_factory=self.remote.search_user,
            query_kwargs={"word": word},
            cache_appender=lambda data, metadata: self.local.append_search_user(word, data, metadata),
            max_page=1,
            force_expiration=cache_strategy == CacheStrategy.FORCE_EXPIRATION,
        )

    def user_illusts_factory(self, user_id: int,
                             cache_strategy: CacheStrategy) -> AsyncGenerator[LazyIllust, None]:
        return mediate_append(
            cache_factory=self.local.user_illusts,
            remote_factory=self.remote.user_illusts,
            query_kwargs={"user_id": user_id},
            cache_appender=lambda data, metadata: self.local.append_user_illusts(user_id, data, metadata),
            max_item=self.conf.pixiv_random_user_illust_max_item,
            max_page=self.conf.pixiv_random_user_illust_max_page,
            force_expiration=cache_strategy == CacheStrategy.FORCE_EXPIRATION,
        )

    def user_bookmarks_factory(self, user_id: int,
                               cache_strategy: CacheStrategy) -> AsyncGenerator[LazyIllust, None]:
        return mediate_append(
            cache_factory=self.local.user_bookmarks,
            remote_factory=self.remote.user_bookmarks,
            query_kwargs={"user_id": user_id},
            cache_appender=lambda data, metadata: self.local.append_user_bookmarks(user_id, data, metadata),
            max_item=self.conf.pixiv_random_bookmark_max_item,
            max_page=self.conf.pixiv_random_bookmark_max_page,
            force_expiration=cache_strategy == CacheStrategy.FORCE_EXPIRATION,
        )

    def recommended_illusts_factory(self, cache_strategy: CacheStrategy) -> AsyncGenerator[LazyIllust, None]:
        return mediate_many(
            cache_factory=self.local.recommended_illusts,
            remote_factory=self.remote.recommended_illusts,
            query_kwargs={},
            cache_invalidator=self.local.invalidate_recommended_illusts,
            cache_appender=lambda data, metadata: self.local.append_recommended_illusts(data, metadata),
            max_item=self.conf.pixiv_random_recommended_illust_max_item,
            max_page=self.conf.pixiv_random_recommended_illust_max_page,
            force_expiration=cache_strategy == CacheStrategy.FORCE_EXPIRATION,
        )

    def related_illusts_factory(self, illust_id: int,
                                cache_strategy: CacheStrategy) -> AsyncGenerator[LazyIllust, None]:
        return mediate_many(
            cache_factory=self.local.related_illusts,
            remote_factory=self.remote.related_illusts,
            query_kwargs={"illust_id": illust_id},
            cache_invalidator=lambda: self.local.invalidate_related_illusts(illust_id),
            cache_appender=lambda data, metadata: self.local.append_related_illusts(illust_id, data, metadata),
            max_item=self.conf.pixiv_random_related_illust_max_item,
            max_page=self.conf.pixiv_random_related_illust_max_page,
            force_expiration=cache_strategy == CacheStrategy.FORCE_EXPIRATION,
        )

    def illust_ranking_factory(self, mode: RankingMode,
                               cache_strategy: CacheStrategy) -> AsyncGenerator[LazyIllust, None]:
        return mediate_many(
            cache_factory=self.local.illust_ranking,
            remote_factory=self.remote.illust_ranking,
            query_kwargs={"mode": mode},
            cache_invalidator=lambda: self.local.invalidate_illust_ranking(mode),
            cache_appender=lambda data, metadata: self.local.append_illust_ranking(mode, data, metadata),
            max_item=self.conf.pixiv_ranking_fetch_item,
            force_expiration=cache_strategy == CacheStrategy.FORCE_EXPIRATION,
        )

    def image_factory(self, illust_id: int, illust: Illust,
                      cache_strategy: CacheStrategy) -> AsyncGenerator[bytes, None]:
        return mediate_single(
            cache_factory=self.local.image,
            remote_factory=self.remote.image,
            query_kwargs={"illust": illust},
            cache_updater=lambda data, metadata: self.local.update_image(illust.id, data, metadata),
            force_expiration=cache_strategy == CacheStrategy.FORCE_EXPIRATION,
        )

    factories = {
        PixivResType.ILLUST_DETAIL: illust_detail_factory,
        PixivResType.USER_DETAIL: user_detail_factory,
        PixivResType.SEARCH_ILLUST: search_illust_factory,
        PixivResType.SEARCH_USER: search_user_factory,
        PixivResType.USER_ILLUSTS: user_illusts_factory,
        PixivResType.USER_BOOKMARKS: user_bookmarks_factory,
        PixivResType.RECOMMENDED_ILLUSTS: recommended_illusts_factory,
        PixivResType.RELATED_ILLUSTS: related_illusts_factory,
        PixivResType.ILLUST_RANKING: illust_ranking_factory,
        PixivResType.IMAGE: image_factory,
    }

    def agen(self, identifier: SharedAgenIdentifier, cache_strategy: CacheStrategy, **kwargs) -> AsyncGenerator[
        Any, None]:
        if identifier.type in self.factories:
            merged_kwargs = identifier.kwargs | kwargs
            # noinspection PyTypeChecker
            return self.factories[identifier.type](self, cache_strategy=cache_strategy, **merged_kwargs)
        else:
            raise ValueError("invalid identifier: " + str(identifier))

    expires_in = {
        PixivResType.ILLUST_DETAIL: timedelta(seconds=context.require(Config).pixiv_illust_detail_cache_expires_in),
        PixivResType.USER_DETAIL: timedelta(seconds=context.require(Config).pixiv_user_detail_cache_expires_in),
        PixivResType.SEARCH_ILLUST: timedelta(seconds=context.require(Config).pixiv_search_illust_cache_expires_in),
        PixivResType.SEARCH_USER: timedelta(seconds=context.require(Config).pixiv_search_user_cache_expires_in),
        PixivResType.USER_ILLUSTS: timedelta(seconds=context.require(Config).pixiv_user_illusts_cache_expires_in),
        PixivResType.USER_BOOKMARKS: timedelta(seconds=context.require(Config).pixiv_user_bookmarks_cache_expires_in),
        PixivResType.RECOMMENDED_ILLUSTS: timedelta(seconds=context.require(Config).pixiv_other_cache_expires_in),
        PixivResType.RELATED_ILLUSTS: timedelta(seconds=context.require(Config).pixiv_related_illusts_cache_expires_in),
        PixivResType.ILLUST_RANKING: timedelta(seconds=context.require(Config).pixiv_illust_ranking_cache_expires_in),
        PixivResType.IMAGE: timedelta(seconds=context.require(Config).pixiv_download_cache_expires_in),
    }

    def calc_expires_time(self, identifier: SharedAgenIdentifier, update_time: datetime) -> datetime:
        if identifier.type in self.factories:
            return update_time + self.expires_in[identifier.type]
        else:
            raise ValueError("invalid identifier: " + str(identifier))

    async def on_agen_next(self, identifier: SharedAgenIdentifier, item: Any):
        if isinstance(item, PixivRepoMetadata) and not self.get_expires_time(identifier):
            expires_time = self.calc_expires_time(identifier, item.update_time)
            self.set_expires_time(identifier, expires_time)


@context.inject
@context.root.register_singleton()
class MediatorPixivRepo:
    _shared_agen_mgr = Inject(PixivSharedAsyncGeneratorManager)
    _local = Inject(LocalPixivRepo)
    _remote = Inject(RemotePixivRepo)

    async def invalidate_cache(self):
        self._shared_agen_mgr.invalidate_all()
        await self._local.invalidate_all()

    async def illust_detail(self, illust_id: int,
                            cache_strategy: CacheStrategy = CacheStrategy.NORMAL) -> AsyncGenerator[Illust, None]:
        logger.info(f"[mediator] illust_detail {illust_id} "
                    f"cache_strategy={cache_strategy.name}")
        with self._shared_agen_mgr.get(SharedAgenIdentifier(PixivResType.ILLUST_DETAIL, illust_id=illust_id),
                                       cache_strategy) as gen:
            async for x in gen:
                if not isinstance(x, PixivRepoMetadata):
                    yield x

    async def user_detail(self, user_id: int,
                          cache_strategy: CacheStrategy = CacheStrategy.NORMAL) -> AsyncGenerator[User, None]:
        logger.info(f"[mediator] user_detail {user_id} "
                    f"cache_strategy={cache_strategy.name}")
        with self._shared_agen_mgr.get(SharedAgenIdentifier(PixivResType.USER_DETAIL, user_id=user_id),
                                       cache_strategy) as gen:
            async for x in gen:
                if not isinstance(x, PixivRepoMetadata):
                    yield x

    async def search_illust(self, word: str,
                            cache_strategy: CacheStrategy = CacheStrategy.NORMAL) -> AsyncGenerator[LazyIllust, None]:
        logger.info(f"[mediator] search_illust {word} "
                    f"cache_strategy={cache_strategy.name}")
        with self._shared_agen_mgr.get(SharedAgenIdentifier(PixivResType.SEARCH_ILLUST, word=word),
                                       cache_strategy) as gen:
            async for x in gen:
                if not isinstance(x, PixivRepoMetadata):
                    yield x

    async def search_user(self, word: str,
                          cache_strategy: CacheStrategy = CacheStrategy.NORMAL) -> AsyncGenerator[User, None]:
        logger.info(f"[mediator] search_user {word} "
                    f"cache_strategy={cache_strategy.name}")
        with self._shared_agen_mgr.get(SharedAgenIdentifier(PixivResType.SEARCH_USER, word=word),
                                       cache_strategy) as gen:
            async for x in gen:
                if not isinstance(x, PixivRepoMetadata):
                    yield x

    async def user_bookmarks(self, user_id: int = 0,
                             cache_strategy: CacheStrategy = CacheStrategy.NORMAL) -> AsyncGenerator[LazyIllust, None]:
        logger.info(f"[mediator] user_bookmarks {user_id} "
                    f"cache_strategy={cache_strategy.name}")
        with self._shared_agen_mgr.get(SharedAgenIdentifier(PixivResType.USER_BOOKMARKS, user_id=user_id),
                                       cache_strategy) as gen:
            async for x in gen:
                if not isinstance(x, PixivRepoMetadata):
                    yield x

    async def user_illusts(self, user_id: int = 0,
                           cache_strategy: CacheStrategy = CacheStrategy.NORMAL) -> AsyncGenerator[LazyIllust, None]:
        logger.info(f"[mediator] user_illusts {user_id} "
                    f"cache_strategy={cache_strategy.name}")
        with self._shared_agen_mgr.get(SharedAgenIdentifier(PixivResType.USER_ILLUSTS, user_id=user_id),
                                       cache_strategy) as gen:
            async for x in gen:
                if not isinstance(x, PixivRepoMetadata):
                    yield x

    async def _contact_user_illusts_with_preview(self, user_preview: UserPreview,
                                                 cache_strategy: CacheStrategy = CacheStrategy.NORMAL) \
            -> AsyncGenerator[LazyIllust, None]:
        for illust in user_preview.illusts:
            yield LazyIllust(illust.id, illust)

        ids = {x.id for x in user_preview.illusts}

        if len(user_preview.illusts) >= 3:
            gen = self.user_illusts(user_preview.user.id, cache_strategy)

            async for x in gen:
                if x.id not in ids:
                    yield x

    async def user_following_illusts(self, user_id: int,
                                     cache_strategy: CacheStrategy = CacheStrategy.NORMAL) \
            -> AsyncGenerator[LazyIllust, None]:
        logger.info(f"[mediator] user_following_illusts {user_id} "
                    f"cache_strategy={cache_strategy.name}")

        n = 0
        gen = []
        peeked = []

        async for user_preview in self._remote.user_following_with_preview(user_id):
            if not isinstance(user_preview, PixivRepoMetadata):
                n += 1
                gen.append(self._contact_user_illusts_with_preview(user_preview, cache_strategy))
                peeked.append(None)

        try:
            while True:
                select = -1
                for i in range(n):
                    try:
                        if not peeked[i]:
                            peeked[i] = await gen[i].__anext__()

                        if select == -1 or peeked[i].create_date > peeked[select].create_date:
                            select = i
                    except StopAsyncIteration:
                        pass

                if select == -1:
                    break

                yield peeked[select]
                peeked[select] = None
        except GeneratorExit:
            for gen in gen:
                await gen.aclose()

    async def recommended_illusts(self, cache_strategy: CacheStrategy = CacheStrategy.NORMAL) \
            -> AsyncGenerator[LazyIllust, None]:
        logger.info(f"[mediator] recommended_illusts "
                    f"cache_strategy={cache_strategy.name}")
        with self._shared_agen_mgr.get(SharedAgenIdentifier(PixivResType.RECOMMENDED_ILLUSTS),
                                       cache_strategy) as gen:
            async for x in gen:
                if not isinstance(x, PixivRepoMetadata):
                    yield x

    async def related_illusts(self, illust_id: int,
                              cache_strategy: CacheStrategy = CacheStrategy.NORMAL) -> AsyncGenerator[LazyIllust, None]:
        logger.info(f"[mediator] related_illusts {illust_id} "
                    f"cache_strategy={cache_strategy.name}")
        with self._shared_agen_mgr.get(SharedAgenIdentifier(PixivResType.RELATED_ILLUSTS, illust_id=illust_id),
                                       cache_strategy) as gen:
            async for x in gen:
                if not isinstance(x, PixivRepoMetadata):
                    yield x

    async def illust_ranking(self, mode: Union[str, RankingMode],
                             cache_strategy: CacheStrategy = CacheStrategy.NORMAL) -> AsyncGenerator[LazyIllust, None]:
        if isinstance(mode, str):
            mode = RankingMode[mode]

        logger.info(f"[mediator] illust_ranking {mode} "
                    f"cache_strategy={cache_strategy.name}")
        with self._shared_agen_mgr.get(SharedAgenIdentifier(PixivResType.ILLUST_RANKING, mode=mode),
                                       cache_strategy) as gen:
            async for x in gen:
                if not isinstance(x, PixivRepoMetadata):
                    yield x

    async def image(self, illust: Illust,
                    cache_strategy: CacheStrategy = CacheStrategy.NORMAL) -> AsyncGenerator[bytes, None]:
        logger.info(f"[mediator] image {illust.id} "
                    f"cache_strategy={cache_strategy.name}")
        with self._shared_agen_mgr.get(SharedAgenIdentifier(PixivResType.IMAGE, illust_id=illust.id),
                                       cache_strategy, illust=illust) as gen:
            async for x in gen:
                if not isinstance(x, PixivRepoMetadata):
                    yield x


__all__ = ('MediatorPixivRepo',)



from datetime import datetime, timezone
from typing import List, AsyncGenerator, TypeVar, Union, Callable, Awaitable, Optional, Any, Mapping

from nonebot import logger

from .errors import NoSuchItemError, CacheExpiredError
from .models import PixivRepoMetadata

T = TypeVar("T")


async def mediate_single(cache_factory: Callable[..., AsyncGenerator[Union[T, PixivRepoMetadata], None]],
                         remote_factory: Callable[..., AsyncGenerator[Union[T, PixivRepoMetadata], None]],
                         query_kwargs: Mapping[str, Any],
                         cache_updater: Callable[[T, Optional[PixivRepoMetadata]], Awaitable[Any]],
                         force_expiration: bool = False) -> AsyncGenerator[Union[T, PixivRepoMetadata], None]:
    try:
        if force_expiration:
            raise NoSuchItemError()
        async for x in cache_factory(**query_kwargs):
            yield x
    except (NoSuchItemError, CacheExpiredError):
        logger.info("[mediator] no cache or cache expired")

        content = None
        metadata = None

        async for x in remote_factory(**query_kwargs):
            if isinstance(x, PixivRepoMetadata):
                metadata = x
            else:
                content = x

        if metadata:
            # 先update再yield，受到SharedAsyncGeneratorManager的影响finally内的语句无法按时执行
            await cache_updater(content, metadata)

            yield metadata
            yield content

        else:
            raise RuntimeError("no metadata")


async def _load_many_from_local_and_remote_and_append(
        cache_factory: Callable[..., AsyncGenerator[Union[T, PixivRepoMetadata], None]],
        remote_factory: Callable[..., AsyncGenerator[Union[T, PixivRepoMetadata], None]],
        query_kwargs: Mapping[str, Any],
        cache_appender: Callable[[List[T], Optional[PixivRepoMetadata]], Awaitable[Any]],
        max_item: int,
        max_page: int):
    loaded_items = 0
    loaded_pages = 0

    # first load from cache
    metadata = None
    async for x in cache_factory(**query_kwargs):
        if isinstance(x, PixivRepoMetadata):
            metadata = x
            loaded_pages = metadata.pages
        else:
            loaded_items += 1

        yield x

        # check whether we approach limit
        if loaded_items >= max_item or loaded_pages >= max_page:
            return

    # then check metadata["next_qs"] and load from remote
    if metadata and metadata.next_qs:
        async for x in _load_many_from_remote_and_append(remote_factory, metadata.next_qs,
                                                         cache_appender,
                                                         max_item - loaded_items, max_page - loaded_pages,
                                                         loaded_pages):
            yield x


async def _load_many_from_remote_and_append(
        remote_factory: Callable[..., AsyncGenerator[Union[T, PixivRepoMetadata], None]],
        query_kwargs: Mapping[str, Any],
        cache_appender: Callable[[T, Optional[PixivRepoMetadata]], Awaitable[Any]],
        max_item: int,
        max_page: int,
        metadata_page_offset: int = 0):
    loaded_items = 0
    loaded_pages = 0

    buffer = []

    async for item in remote_factory(**query_kwargs):
        if isinstance(item, PixivRepoMetadata):
            item.pages += metadata_page_offset
            loaded_pages = item.pages

            if len(buffer) > 0:
                await cache_appender(buffer, item)

                for x in buffer:
                    yield x

                buffer.clear()

                # check whether we approach limit
                if loaded_items >= max_item or loaded_pages >= max_page:
                    return

            yield item
        else:
            loaded_items += 1
            buffer.append(item)


async def mediate_many(cache_factory: Callable[..., AsyncGenerator[Union[T, PixivRepoMetadata], None]],
                       remote_factory: Callable[..., AsyncGenerator[Union[T, PixivRepoMetadata], None]],
                       query_kwargs: Mapping[str, Any],
                       cache_invalidator: Callable[[], Awaitable[Any]],
                       cache_appender: Callable[[List[T], Optional[PixivRepoMetadata]], Awaitable[Any]],
                       max_item: int = 2 ** 31,
                       max_page: int = 2 ** 31,
                       force_expiration: bool = False) -> AsyncGenerator[Union[T, PixivRepoMetadata], None]:
    try:
        if force_expiration:
            raise NoSuchItemError()

        async for x in _load_many_from_local_and_remote_and_append(cache_factory, remote_factory, query_kwargs,
                                                                   cache_appender, max_item, max_page):
            yield x
    except NoSuchItemError:
        logger.info("[mediator] no cache")
        async for x in _load_many_from_remote_and_append(remote_factory, query_kwargs,
                                                         cache_appender,
                                                         max_item, max_page):
            yield x
    except CacheExpiredError:
        logger.info("[mediator] cache expired")
        await cache_invalidator()
        async for x in _load_many_from_remote_and_append(remote_factory, query_kwargs,
                                                         cache_appender,
                                                         max_item, max_page):
            yield x


async def mediate_append(cache_factory: Callable[..., AsyncGenerator[Union[T, PixivRepoMetadata], None]],
                         remote_factory: Callable[..., AsyncGenerator[Union[T, PixivRepoMetadata], None]],
                         query_kwargs: Mapping[str, Any],
                         cache_appender: Callable[[List[T], Optional[PixivRepoMetadata]], Awaitable[bool]],
                         max_item: int = 2 ** 31,
                         max_page: int = 2 ** 31,
                         force_expiration: bool = False) -> AsyncGenerator[Union[T, PixivRepoMetadata], None]:
    try:
        if force_expiration:
            metadata = None
            async for x in cache_factory(**query_kwargs):
                if isinstance(x, PixivRepoMetadata):
                    metadata = x
                    break

            if not metadata:
                raise NoSuchItemError()
            else:
                raise CacheExpiredError(metadata)

        async for x in _load_many_from_local_and_remote_and_append(cache_factory, remote_factory, query_kwargs,
                                                                   cache_appender, max_item, max_page):
            yield x
    except NoSuchItemError:
        logger.info("[mediator] no cache")
        async for x in _load_many_from_remote_and_append(remote_factory, query_kwargs,
                                                         cache_appender,
                                                         max_item, max_page):
            yield x
    except CacheExpiredError as e:
        logger.info("[mediator] cache expired")
        loaded_items = 0
        loaded_pages = 0

        # peek new bookmarks from remote
        # then append them to local cache until we found some items already exist
        buffer = []
        metadata = e.metadata

        async for x in remote_factory(**query_kwargs):
            if isinstance(x, PixivRepoMetadata):
                loaded_pages = x.pages
                # we don't use this metadata

                if len(buffer) > 0:
                    metadata.update_time = datetime.now(timezone.utc)
                    if await cache_appender(buffer, metadata):
                        break
                    buffer = []

                    # check whether we approach limit
                    if loaded_items >= max_item or loaded_pages >= max_page:
                        return
            else:
                loaded_items += 1
                buffer.append(x)

        async for x in _load_many_from_local_and_remote_and_append(cache_factory, remote_factory, query_kwargs,
                                                                   cache_appender, max_item, max_page):
            yield x


__all__ = ("mediate_single", "mediate_many", "mediate_append")



from typing import Callable

from nonebot_plugin_pixivbot import context
from nonebot_plugin_pixivbot.config import Config
from nonebot_plugin_pixivbot.context import Inject
from nonebot_plugin_pixivbot.data.source.sql import SqlDataSource
from nonebot_plugin_pixivbot.enums import DataSourceType
from nonebot_plugin_pixivbot.handler.interceptor.base import Interceptor, UID, GID
from nonebot_plugin_pixivbot.protocol_dep.post_dest import PostDestination


@context.inject
@context.register_singleton()
class SqlRemoveSessionInterceptor(Interceptor):
    conf: Config = Inject(Config)

    async def intercept(self, wrapped_func: Callable, *args,
                        post_dest: PostDestination[UID, GID],
                        silently: bool, **kwargs):
        await wrapped_func(*args, post_dest=post_dest, silently=silently, **kwargs)

        if self.conf.pixiv_data_source == DataSourceType.sqlite:
            await context.require(SqlDataSource).session.remove()



from functools import partial
from typing import Callable, TypeVar, Type, Optional, Iterable

from nonebot_plugin_pixivbot.protocol_dep.post_dest import PostDestination
from .base import Interceptor

UID = TypeVar("UID")
GID = TypeVar("GID")


class CombinedInterceptor(Interceptor):
    def __init__(self, x: Interceptor, y: Interceptor):
        self.x = x
        self.y = y

    @staticmethod
    def from_iterable(interceptors: Iterable[Interceptor]) -> "Interceptor":
        itcp = None
        for x in interceptors:
            if itcp is None:
                itcp = x
            else:
                itcp = CombinedInterceptor(itcp, x)
        if itcp is None:
            raise ValueError("interceptors has no element")
        return itcp

    def find(self, interceptor_type: Type[Interceptor]) -> Optional[Interceptor]:
        if isinstance(interceptor_type, CombinedInterceptor):
            raise ValueError("you are attempting to find a CombinedInterceptor")
        if isinstance(self.x, interceptor_type):
            return self.x
        elif isinstance(self.y, interceptor_type):
            return self.y
        else:
            result = None
            if isinstance(self.x, CombinedInterceptor):
                result = self.x.find(interceptor_type)
            if not result and isinstance(self.y, CombinedInterceptor):
                result = self.y.find(interceptor_type)
            return result

    async def intercept(self, wrapped_func: Callable, *args,
                        post_dest: PostDestination[UID, GID],
                        silently: bool,
                        **kwargs):
        await self.x.intercept(
            partial(self.y.intercept, wrapped_func), *args,
            post_dest=post_dest, silently=silently, **kwargs
        )



from typing import Protocol

from nonebot_plugin_pixivbot import context
from nonebot_plugin_pixivbot.config import Config
from nonebot_plugin_pixivbot.enums import DataSourceType


class SeqRepo(Protocol):
    async def inc_and_get(self, key: str) -> int:
        ...

    async def get_and_inc(self, key: str) -> int:
        ...

    async def get(self, key: str) -> int:
        ...


conf = context.require(Config)
if conf.pixiv_data_source == DataSourceType.mongo:
    from .mongo import MongoSeqRepo

    context.bind(SeqRepo, MongoSeqRepo)
else:
    from .sql import SqlSeqRepo

    context.bind(SeqRepo, SqlSeqRepo)

__all__ = ("SeqRepo",)



"""
nonebot-plugin-pixivbot

@Author         : ssttkkl
@License        : MIT
@GitHub         : https://github.com/ssttkkl/nonebot-plugin-pixivbot
"""

help_text = """
常规语句：
- 看看榜<范围>：查看pixiv榜单
- 来张图：从推荐插画随机抽选一张插画
- 来张<关键字>图：搜索关键字，从搜索结果随机抽选一张插画
- 来张<用户>老师的图：搜索画师，从该画师的插画列表里随机抽选一张插画
- 看看图<插画ID>：查看id为<插画ID>的插画
- 来张私家车：从书签中随机抽选一张插画
- 还要：重复上一次请求
- 不够色：获取上一张插画的相关推荐

命令语句：
- /pixivbot help：查看本帮助
- /pixivbot bind：绑定Pixiv账号

更多功能：参见https://github.com/ssttkkl/nonebot-plugin-pixivbot
""".strip()

from nonebot.plugin import PluginMetadata

__plugin_meta__ = PluginMetadata(
    name='PixivBot',
    description='发送随机Pixiv插画、画师更新推送、定时订阅推送……',
    usage=help_text,
    extra={'version': '1.4.0'}
)

# ================= provide beans =================
from .global_context import context
from .nb_providers import provide

provide(context)

# =========== register handler & service ============
from . import handler
from . import service

# ============== load custom protocol =============
from importlib import import_module
from nonebot import logger

supported_modules = ["nonebot_plugin_pixivbot_onebot_v11", "nonebot_plugin_pixivbot_kook"]

for p in supported_modules:
    try:
        import_module(p)
        logger.success("Loaded Module: " + p)
    except ModuleNotFoundError:
        pass

__all__ = ("context", "__plugin_meta__")



from abc import ABC, abstractmethod
from contextvars import ContextVar
from typing import Generic, TypeVar, Optional

from nonebot import Bot
from nonebot.internal.adapter import Event

from nonebot_plugin_pixivbot.global_context import context
from nonebot_plugin_pixivbot.model.identifier import PostIdentifier
from nonebot_plugin_pixivbot.protocol_dep.protocol_dep import ProtocolDep, ProtocolDepManager
from nonebot_plugin_pixivbot.utils.nonebot import get_adapter_name

UID = TypeVar("UID")
GID = TypeVar("GID")


class PostDestination(ABC, Generic[UID, GID]):

    @property
    @abstractmethod
    def identifier(self) -> PostIdentifier[UID, GID]:
        raise NotImplementedError()

    @property
    def adapter(self) -> str:
        return self.identifier.adapter

    @property
    def user_id(self) -> Optional[UID]:
        return self.identifier.user_id

    @property
    def group_id(self) -> Optional[GID]:
        return self.identifier.group_id

    def __str__(self) -> str:
        return str(self.identifier)

    @abstractmethod
    def normalized(self) -> "PostDestination[UID, GID]":
        """
        返回一个不含任何附加信息（如引用消息）的PostDestination
        :return:
        """
        raise NotImplementedError()


current_post_dest: ContextVar[PostDestination[UID, GID]] = ContextVar("current_post_dest")


class PostDestinationFactory(ProtocolDep, ABC, Generic[UID, GID]):
    @abstractmethod
    def build(self, bot: Bot, user_id: Optional[UID], group_id: Optional[GID]) -> PostDestination:
        raise NotImplementedError()

    @abstractmethod
    def from_event(self, bot: Bot, event: Event) -> PostDestination:
        raise NotImplementedError()


@context.register_singleton()
class PostDestinationFactoryManager(ProtocolDepManager[PostDestinationFactory]):
    def build(self, bot: Bot, user_id: Optional[UID], group_id: Optional[GID]) -> PostDestination[UID, GID]:
        return self[get_adapter_name(bot)].build(bot, user_id, group_id)

    def from_event(self, bot: Bot, event: Event) -> PostDestination[UID, GID]:
        return self[get_adapter_name(bot)].from_event(bot, event)


__all__ = ("PostDestination", "PostDestinationFactory", "PostDestinationFactoryManager",
           "current_post_dest")



from abc import ABC, abstractmethod
from inspect import isawaitable
from typing import Awaitable, Union, TypeVar, Sequence, Optional

from nonebot_plugin_pixivbot.config import Config
from nonebot_plugin_pixivbot.model import Illust
from nonebot_plugin_pixivbot.model.message import IllustMessageModel, IllustMessagesModel
from nonebot_plugin_pixivbot.protocol_dep.post_dest import PostDestination
from .interceptor.base import Interceptor
from .interceptor.combined_interceptor import CombinedInterceptor
from .pkg_context import context
from ..context import Inject
from ..protocol_dep.postman import PostmanManager

UID = TypeVar("UID")
GID = TypeVar("GID")

PD = PostDestination[UID, GID]


@context.inject
class Handler(ABC):
    conf = Inject(Config)
    postman_manager = Inject(PostmanManager)

    def __init__(self):
        self.interceptor = None

    async def post_plain_text(self, message: str,
                              post_dest: PostDestination):
        await self.postman_manager.send_plain_text(message, post_dest=post_dest)

    async def post_illust(self, illust: Illust, *,
                          header: Optional[str] = None,
                          number: Optional[int] = None,
                          post_dest: PD):
        model = await IllustMessageModel.from_illust(illust, header=header, number=number)
        if model is not None:
            await self.postman_manager.send_illust(model, post_dest=post_dest)

    async def post_illusts(self, illusts: Sequence[Illust], *,
                           header: Optional[str] = None,
                           number: Optional[int] = None,
                           post_dest: PD):
        model = await IllustMessagesModel.from_illusts(illusts, header=header, number=number)
        if model:
            await self.postman_manager.send_illusts(model, post_dest=post_dest)

    @classmethod
    @abstractmethod
    def type(cls) -> str:
        raise NotImplementedError()

    @abstractmethod
    def enabled(self) -> bool:
        raise NotImplementedError()

    def parse_args(self, args: Sequence[str], post_dest: PD) -> Union[dict, Awaitable[dict]]:
        """
        将位置参数转化为命名参数
        :param args: 位置参数sequence
        :param post_dest: PostDestination
        :return: 命名参数dict
        """
        return {}

    async def handle(self, *args,
                     post_dest: PD,
                     silently: bool = False,
                     disabled_interceptors: bool = False,
                     **kwargs):
        if not self.enabled():
            return

        if self.interceptor is not None and not disabled_interceptors:
            await self.interceptor.intercept(self._parse_args_and_actual_handle, *args,
                                             post_dest=post_dest,
                                             silently=silently,
                                             **kwargs)
        else:
            await self._parse_args_and_actual_handle(*args, post_dest=post_dest, silently=silently, **kwargs)

    async def handle_with_parsed_args(self, *, post_dest: PD,
                                      silently: bool = False,
                                      disabled_interceptors: bool = False,
                                      **kwargs):
        if not self.enabled():
            return

        if self.interceptor is not None and not disabled_interceptors:
            await self.interceptor.intercept(self.actual_handle,
                                             post_dest=post_dest,
                                             silently=silently,
                                             **kwargs)
        else:
            await self.actual_handle(post_dest=post_dest, silently=silently, **kwargs)

    async def _parse_args_and_actual_handle(self, *args,
                                            post_dest: PD,
                                            silently: bool = False,
                                            **kwargs):
        parsed_kwargs = self.parse_args(args, post_dest)
        if isawaitable(parsed_kwargs):
            parsed_kwargs = await parsed_kwargs

        kwargs = {**kwargs, **parsed_kwargs}
        await self.actual_handle(post_dest=post_dest, silently=silently, **kwargs)

    @abstractmethod
    async def actual_handle(self, *, post_dest: PD,
                            silently: bool = False,
                            **kwargs):
        """
        处理指令
        :param post_dest: PostDestination
        :param silently: 失败时不发送消息
        :param kwargs: 参数dict
        """
        raise NotImplementedError()

    def add_interceptor(self, interceptor: Interceptor):
        if self.interceptor:
            self.interceptor = CombinedInterceptor(
                self.interceptor, interceptor)
        else:
            self.interceptor = interceptor



from typing import List

from beanie import Document
from pymongo import IndexModel

from nonebot_plugin_pixivbot import context
from nonebot_plugin_pixivbot.config import Config
from nonebot_plugin_pixivbot.data.pixiv_repo.models import PixivRepoMetadata
from nonebot_plugin_pixivbot.data.source import MongoDataSource
from nonebot_plugin_pixivbot.enums import RankingMode
from nonebot_plugin_pixivbot.model import Illust, User

conf = context.require(Config)


class PixivRepoCache(Document):
    metadata: PixivRepoMetadata


class IllustSetCache(PixivRepoCache):
    illust_id: List[int]


class UserSetCache(PixivRepoCache):
    user_id: List[int]


class DownloadCache(PixivRepoCache):
    illust_id: int
    content: bytes

    class Settings:
        name = "download_cache"
        indexes = [
            IndexModel([("illust_id", 1)], unique=True),
            IndexModel([("metadata.update_time", 1)], expireAfterSeconds=conf.pixiv_download_cache_expires_in)
        ]


class IllustDetailCache(PixivRepoCache):
    illust: Illust

    class Settings:
        name = "illust_detail_cache"
        indexes = [
            IndexModel([("illust.id", 1)], unique=True),
            IndexModel([("metadata.update_time", 1)], expireAfterSeconds=conf.pixiv_illust_detail_cache_expires_in)
        ]


class IllustRankingCache(IllustSetCache):
    mode: RankingMode

    class Settings:
        name = "illust_ranking_cache"
        indexes = [
            IndexModel([("mode", 1)], unique=True),
            IndexModel([("metadata.update_time", 1)], expireAfterSeconds=conf.pixiv_illust_ranking_cache_expires_in)
        ]


class OtherIllustCache(IllustSetCache):
    type: str

    class Settings:
        name = "other_cache"
        indexes = [
            IndexModel([("type", 1)], unique=True),
            IndexModel([("metadata.update_time", 1)], expireAfterSeconds=conf.pixiv_other_cache_expires_in)
        ]


class RelatedIllustsCache(IllustSetCache):
    original_illust_id: int

    class Settings:
        name = "related_illusts_cache"
        indexes = [
            IndexModel([("original_illust_id", 1)], unique=True),
            IndexModel([("metadata.update_time", 1)], expireAfterSeconds=conf.pixiv_related_illusts_cache_expires_in)
        ]


class SearchIllustCache(IllustSetCache):
    word: str

    class Settings:
        name = "search_illust_cache"
        indexes = [
            IndexModel([("word", 1)], unique=True),
            IndexModel([("metadata.update_time", 1)], expireAfterSeconds=conf.pixiv_search_illust_cache_delete_in)
        ]


class SearchUserCache(UserSetCache):
    word: str
    user_id: List[int]

    class Settings:
        name = "search_user_cache"
        indexes = [
            IndexModel([("word", 1)], unique=True),
            IndexModel([("metadata.update_time", 1)], expireAfterSeconds=conf.pixiv_search_user_cache_delete_in)
        ]


class UserBookmarksCache(IllustSetCache):
    user_id: int

    class Settings:
        name = "user_bookmarks_cache"
        indexes = [
            IndexModel([("user_id", 1)], unique=True),
            IndexModel([("metadata.update_time", 1)], expireAfterSeconds=conf.pixiv_user_bookmarks_cache_delete_in)
        ]


class UserDetailCache(PixivRepoCache):
    user: User

    class Settings:
        name = "user_detail_cache"
        indexes = [
            IndexModel([("user.id", 1)], unique=True),
            IndexModel([("metadata.update_time", 1)], expireAfterSeconds=conf.pixiv_user_detail_cache_expires_in)
        ]


class UserIllustsCache(IllustSetCache):
    user_id: int

    class Settings:
        name = "user_illusts_cache"
        indexes = [
            IndexModel([("user_id", 1)], unique=True),
            IndexModel([("metadata.update_time", 1)], expireAfterSeconds=conf.pixiv_user_illusts_cache_delete_in)
        ]


context.require(MongoDataSource).document_models.extend([
    DownloadCache, IllustDetailCache, IllustRankingCache, OtherIllustCache,
    RelatedIllustsCache, SearchIllustCache, SearchUserCache,
    UserBookmarksCache, UserDetailCache, UserIllustsCache
])

__all__ = ("PixivRepoCache", "IllustSetCache",
           "DownloadCache", "IllustDetailCache", "IllustRankingCache", "OtherIllustCache",
           "RelatedIllustsCache", "SearchIllustCache", "SearchUserCache",
           "UserBookmarksCache", "UserDetailCache", "UserIllustsCache")



from nonebot_plugin_pixivbot import context

from .base import PixivRepo
from .lazy_illust import LazyIllust
from .mediator_repo import MediatorPixivRepo

context.bind(PixivRepo, MediatorPixivRepo)

__all__ = ("PixivRepo", "LazyIllust",)



from typing import TypeVar, Optional

from nonebot_plugin_pixivbot.context import Inject
from nonebot_plugin_pixivbot.data.pixiv_binding import PixivBindingRepo
from nonebot_plugin_pixivbot.global_context import context
from nonebot_plugin_pixivbot.model.pixiv_binding import PixivBinding

UID = TypeVar("UID")


@context.inject
@context.register_singleton()
class PixivAccountBinder:
    repo = Inject(PixivBindingRepo)

    async def bind(self, adapter: str, user_id: UID, pixiv_user_id: int):
        binding = PixivBinding(adapter=adapter, user_id=user_id, pixiv_user_id=pixiv_user_id)
        await self.repo.update(binding)

    async def unbind(self, adapter: str, user_id: UID) -> bool:
        return await self.repo.remove(adapter, user_id)

    async def get_binding(self, adapter: str, user_id: UID) -> Optional[int]:
        binding = await self.repo.get(adapter, user_id)
        if binding:
            return binding.pixiv_user_id
        else:
            return None



from typing import Union, AsyncGenerator, Protocol

from nonebot_plugin_pixivbot.enums import RankingMode
from nonebot_plugin_pixivbot.model import Illust, User
from .lazy_illust import LazyIllust
from .models import PixivRepoMetadata


class PixivRepo(Protocol):
    def illust_detail(self, illust_id: int) -> AsyncGenerator[Union[Illust, PixivRepoMetadata], None]:
        ...

    def user_detail(self, user_id: int) -> AsyncGenerator[Union[User, PixivRepoMetadata], None]:
        ...

    def search_illust(self, word: str) -> AsyncGenerator[Union[LazyIllust, PixivRepoMetadata], None]:
        ...

    def search_user(self, word: str) -> AsyncGenerator[Union[User, PixivRepoMetadata], None]:
        ...

    def user_illusts(self, user_id: int) -> AsyncGenerator[Union[LazyIllust, PixivRepoMetadata], None]:
        ...

    def user_bookmarks(self, user_id: int = 0) -> AsyncGenerator[Union[LazyIllust, PixivRepoMetadata], None]:
        ...

    def recommended_illusts(self) -> AsyncGenerator[Union[LazyIllust, PixivRepoMetadata], None]:
        ...

    def related_illusts(self, illust_id: int) -> AsyncGenerator[Union[LazyIllust, PixivRepoMetadata], None]:
        ...

    def illust_ranking(self, mode: Union[str, RankingMode]) \
            -> AsyncGenerator[Union[LazyIllust, PixivRepoMetadata], None]:
        ...

    def image(self, illust: Illust) -> AsyncGenerator[Union[bytes, PixivRepoMetadata], None]:
        ...


__all__ = ("PixivRepo",)



from .identifier import UserIdentifier, PostIdentifier
from .illust import Illust
from .pixiv_binding import PixivBinding
from .subscription import Subscription, ScheduleType
from .tag import Tag
from .user import User
from .user_preview import UserPreview
from .watch_task import WatchTask, WatchType

__all__ = ("Illust", "User", "UserPreview", "Tag", "Subscription", "ScheduleType",
           "UserIdentifier", "PostIdentifier", "PixivBinding",
           "WatchTask", "WatchType")



from enum import Enum


class PixivResType(Enum):
    SEARCH_ILLUST = 0
    SEARCH_USER = 1
    USER_ILLUSTS = 2
    USER_BOOKMARKS = 3
    RECOMMENDED_ILLUSTS = 4
    ILLUST_RANKING = 5
    ILLUST_DETAIL = 6
    IMAGE = 7
    RELATED_ILLUSTS = 8
    USER_DETAIL = 9


class CacheStrategy(Enum):
    NORMAL = 0
    FORCE_EXPIRATION = 1



from typing import TypeVar, Sequence

from nonebot_plugin_pixivbot.context import Inject
from nonebot_plugin_pixivbot.protocol_dep.post_dest import PostDestination
from nonebot_plugin_pixivbot.service.pixiv_account_binder import PixivAccountBinder
from nonebot_plugin_pixivbot.utils.errors import BadRequestError
from .command import SubCommandHandler, CommandHandler
from ..pkg_context import context

UID = TypeVar("UID")
GID = TypeVar("GID")


@context.inject
@context.require(CommandHandler).sub_command("bind")
class BindHandler(SubCommandHandler):
    binder = Inject(PixivAccountBinder)

    @classmethod
    def type(cls) -> str:
        return "bind"

    def enabled(self) -> bool:
        return True

    async def parse_args(self, args: Sequence[str], post_dest: PostDestination[UID, GID]) -> dict:
        if len(args) < 1:
            raise BadRequestError()
        else:
            try:
                pixiv_user_id = int(args[0])
            except ValueError as e:
                raise BadRequestError("请输入正确格式的Pixiv账号") from e

            return {"pixiv_user_id": pixiv_user_id}

    # noinspection PyMethodOverriding
    async def actual_handle(self, *, pixiv_user_id: int,
                            post_dest: PostDestination[UID, GID],
                            silently: bool = False):
        await self.binder.bind(post_dest.adapter, post_dest.user_id, pixiv_user_id)
        await self.post_plain_text(message="Pixiv账号绑定成功", post_dest=post_dest)

    async def actual_handle_bad_request(self, err: BadRequestError,
                                        *, post_dest: PostDestination[UID, GID],
                                        silently: bool = False):
        if not silently:
            if err.message:
                await self.post_plain_text(message=err.message, post_dest=post_dest)
            else:
                pixiv_user_id = await self.binder.get_binding(post_dest.adapter, post_dest.user_id)
                if pixiv_user_id is not None:
                    msg = f"当前绑定账号：{pixiv_user_id}\n"
                else:
                    msg = "当前未绑定Pixiv账号\n"
                msg += "命令格式：/pixivbot bind <pixiv_user_id>"
                await self.post_plain_text(message=msg, post_dest=post_dest)


@context.inject
@context.require(CommandHandler).sub_command("unbind")
class UnbindHandler(SubCommandHandler):
    binder = Inject(PixivAccountBinder)

    @classmethod
    def type(cls) -> str:
        return "unbind"

    def enabled(self) -> bool:
        return True

    def parse_args(self, args: Sequence[str], post_dest: PostDestination[UID, GID]) -> dict:
        return {}

    # noinspection PyMethodOverriding
    async def actual_handle(self, *, post_dest: PostDestination[UID, GID],
                            silently: bool = False):
        result = await self.binder.unbind(post_dest.adapter, post_dest.user_id)
        if result:
            await self.post_plain_text(message="Pixiv账号解绑成功", post_dest=post_dest)
        else:
            await self.post_plain_text(message="当前未绑定Pixiv账号", post_dest=post_dest)



from abc import ABC
from typing import Type, Callable, TypeVar, Sequence

from lazy import lazy
from nonebot import Bot, on_command
from nonebot import logger
from nonebot.internal.adapter import Event
from nonebot.internal.matcher import Matcher
from nonebot.internal.params import Depends
from nonebot.typing import T_State

from nonebot_plugin_pixivbot.global_context import context
from nonebot_plugin_pixivbot.protocol_dep.post_dest import PostDestination
from nonebot_plugin_pixivbot.utils.errors import BadRequestError
from ..entry_handler import EntryHandler
from ..entry_handler import post_destination
from ..handler import Handler
from ..utils import get_command_rule

UID = TypeVar("UID")
GID = TypeVar("GID")


class SubCommandHandler(Handler, ABC):
    async def handle(self, *args,
                     post_dest: PostDestination[UID, GID],
                     silently: bool = False,
                     disabled_interceptors: bool = False,
                     **kwargs):
        try:
            await super().handle(*args, post_dest=post_dest, silently=silently,
                                 disabled_interceptors=disabled_interceptors, **kwargs)
        except BadRequestError as e:
            await self.handle_bad_request(err=e, post_dest=post_dest, silently=silently)

    async def handle_bad_request(self, err: BadRequestError, *,
                                 post_dest: PostDestination[UID, GID],
                                 silently: bool = False):
        if self.interceptor is not None:
            await self.interceptor.intercept(self.actual_handle_bad_request, err,
                                             post_dest=post_dest, silently=silently)
        else:
            await self.actual_handle_bad_request(err, post_dest=post_dest, silently=silently)

    async def actual_handle_bad_request(self, err: BadRequestError, *,
                                        post_dest: PostDestination[UID, GID],
                                        silently: bool = False):
        if not silently:
            await self.post_plain_text(err.message, post_dest=post_dest)


@context.root.register_eager_singleton()
class CommandHandler(EntryHandler):
    def __init__(self):
        super().__init__()
        self.handlers = dict[str, Type[SubCommandHandler]]()

    @classmethod
    def type(cls) -> str:
        return "command"

    def enabled(self) -> bool:
        return True

    @lazy
    def matcher(self):
        return on_command("pixivbot", rule=get_command_rule(), priority=5)

    async def on_match(self, bot: Bot, event: Event, state: T_State, matcher: Matcher,
                       post_dest: PostDestination[UID, GID] = Depends(post_destination)):
        args = str(event.get_message()).strip().split()[1:]
        await self.handle(*args, post_dest=post_dest)

    def sub_command(self, type: str) \
            -> Callable[[Type[SubCommandHandler]], Type[SubCommandHandler]]:
        def decorator(cls: Type[SubCommandHandler]):
            if cls not in context:
                context.root.register_singleton()(cls)
            self.handlers[type] = cls
            logger.trace(f"registered subcommand {type}")
            return cls

        return decorator

    def parse_args(self, args: Sequence[str], post_dest: PostDestination[UID, GID]) -> dict:
        return {"args": args}

    # noinspection PyMethodOverriding
    async def actual_handle(self, *, args: Sequence[str],
                            post_dest: PostDestination[UID, GID],
                            silently: bool = False):
        logger.debug("args: " + " ".join(map(str, args)))
        if len(args) == 0:
            handler = context.require(self.handlers["help"])
        elif args[0] not in self.handlers:
            if not silently:
                await self.post_plain_text(f"不存在命令 '{args[0]}'", post_dest=post_dest)
            return
        else:
            handler = context.require(self.handlers[args[0]])

        await handler.handle(*args[1:], post_dest=post_dest)



from typing import Union, Awaitable, TypeVar, Sequence

from nonebot_plugin_pixivbot import help_text
from nonebot_plugin_pixivbot.protocol_dep.post_dest import PostDestination
from .command import SubCommandHandler, CommandHandler
from ..pkg_context import context

UID = TypeVar("UID")
GID = TypeVar("GID")


@context.require(CommandHandler).sub_command("help")
class HelpHandler(SubCommandHandler):
    @classmethod
    def type(cls) -> str:
        return "help"

    def enabled(self) -> bool:
        return True

    def parse_args(self, args: Sequence[str], post_dest: PostDestination[UID, GID]) \
            -> Union[dict, Awaitable[dict]]:
        return {}

    async def actual_handle(self, *, post_dest: PostDestination[UID, GID],
                            silently: bool = False, **kwargs):
        await self.post_plain_text(help_text, post_dest=post_dest)



from typing import TypeVar, AsyncGenerator, Optional, Any, Collection

from beanie import Document, BulkWriter
from pymongo import IndexModel
from pymongo.errors import DuplicateKeyError

from nonebot_plugin_pixivbot.context import Inject
from nonebot_plugin_pixivbot.global_context import context
from nonebot_plugin_pixivbot.model import WatchTask, PostIdentifier
from ..seq import SeqRepo
from ..source import MongoDataSource
from ..utils.process_subscriber import process_subscriber

UID = TypeVar("UID")
GID = TypeVar("GID")

ID = PostIdentifier[UID, GID]


class WatchTaskDocument(WatchTask[Any, Any], Document):
    class Settings:
        name = "watch_task"
        indexes = [
            IndexModel([("subscriber.adapter", 1)]),
            IndexModel([("subscriber", 1), ("code", 1)]),
            IndexModel([("subscriber", 1), ("type", 1), ("kwargs", 1)], unique=True)
        ]


context.require(MongoDataSource).document_models.append(WatchTaskDocument)


@context.inject
@context.register_singleton()
class MongoWatchTaskRepo:
    mongo = Inject(MongoDataSource)
    seq_repo: SeqRepo = Inject(SeqRepo)

    async def get_by_subscriber(self, subscriber: ID) -> AsyncGenerator[WatchTask, None]:
        subscriber = process_subscriber(subscriber)
        async for doc in WatchTaskDocument.find(WatchTaskDocument.subscriber == subscriber):
            yield doc

    async def get_by_adapter(self, adapter: str) -> AsyncGenerator[WatchTask, None]:
        async for doc in WatchTaskDocument.find(WatchTaskDocument.subscriber.adapter == adapter):
            yield doc

    async def get_by_code(self, subscriber: ID, code: int) -> Optional[WatchTask]:
        subscriber = process_subscriber(subscriber)
        return await WatchTaskDocument.find_one(WatchTaskDocument.subscriber == subscriber,
                                                WatchTaskDocument.code == code)

    async def insert(self, task: WatchTask) -> bool:
        try:
            task.subscriber = process_subscriber(task.subscriber)
            doc = WatchTaskDocument(**task.dict())
            await doc.save()

            task.code = await self.seq_repo.inc_and_get(f'watch_task {task.subscriber}')
            doc.code = task.code
            await doc.save()

            return True
        except DuplicateKeyError:
            return False

    async def update(self, task: WatchTask) -> bool:
        if isinstance(task, WatchTaskDocument):
            await task.save()
        else:
            task.subscriber = process_subscriber(task.subscriber)
            await WatchTaskDocument.find_one(
                WatchTaskDocument.subscriber == task.subscriber,
                WatchTaskDocument.code == task.code
            ).update(**task.dict(exclude={"subscriber", "code"}))
        return True

    async def delete_one(self, subscriber: ID, code: int) -> Optional[WatchTask]:
        # beanie不支持原子性的find_one_and_delete操作
        subscriber = process_subscriber(subscriber)
        query = {
            "code": code,
            "subscriber": process_subscriber(subscriber).dict()
        }
        result = await self.mongo.db.watch_task.find_one_and_delete(query)
        if result:
            return WatchTask.parse_obj(result)
        else:
            return None

    async def delete_many_by_subscriber(self, subscriber: ID) -> Collection[WatchTask]:
        subscriber = process_subscriber(subscriber)

        old_doc = await WatchTaskDocument.find(
            WatchTaskDocument.subscriber == subscriber
        ).to_list()

        async with BulkWriter() as bw:
            for x in old_doc:
                await x.delete(bulk_writer=bw)

        return old_doc


__all__ = ("MongoWatchTaskRepo",)
